"""
Run a reduction workflow.

:func:`process_template` evaluates a template for the given input values.

:func:`find_calculated` returns the list the template nodes that have already
been calculated and cached for the given input values.

:func:`fingerprint_template` returns the unique fingerprint for each node
in the template given its input values.
"""
from __future__ import print_function

import sys

import hashlib
import contextlib
from inspect import getsource

from .anno_exc import annotate_exception
from .cache import get_cache
from .core import lookup_module, lookup_datatype
from .core import Bundle
from .automod import validate

IS_PY3 = sys.version_info[0] >= 3

def find_calculated(template, config):
    """
    Returns a boolean vector indicating whether or not each node in the
    template has been computed and cached.
    """
    cache = get_cache()

    fingerprints = fingerprint_template(template, config)
    return [cache.exists(fingerprints[node])
            for node, _ in enumerate(template.modules)]


def process_template(template, config, target=(None, None)):
    """
    Evaluate the template.

    If *target=(node number, "terminal id")* is specified, then only
    calculate the nodes required to evaluate the target.

    For each node, if its fingerprint is already in the cache, retrieve
    the cached value and use it for subsequent nodes.  If not, then run
    the node action and place the results in the cache.

    If *target* is specified, then return the target as a json serialized
    object containing the list of values on the specified output terminal.
    """
    cache = get_cache()

    results = {}
    return_node, return_terminal = target

    fingerprints = fingerprint_template(template, config)
    for node, input_wires in template.ordered(target=return_node):
        node_info = template.modules[node]
        module = lookup_module(node_info['module'])
        node_id = "node %d, %s"%(node, node_info['module'])
        input_terminals = module.inputs
        module.outputs = module.outputs

        # Build the inputs; if returning an input terminal, put it in the
        # results set.   This is extra work for the case where the results
        # have already been computed, but it simplifies the code for the
        # case where the return target is an input terminal.
        inputs = _get_inputs(results, input_wires, input_terminals)
        if return_node == node and return_terminal in inputs:
            # We are returning inputs, so treat them as if it were outputs.
            # That means putting them into a bundle so that we can convert
            # to and from JSON.  But we have to find the terminal first so
            # that we know the datatype.  Since we are only returning the
            # inputs, we don't need to compute the node outputs, and we
            # can return immediately.
            for terminal in input_terminals:
                if terminal["id"] == return_terminal:
                    return _bundle(terminal, inputs[return_terminal])

        # If the module has been flagged "nocache" for debugging, then
        # clear all cached entries that depend on it.  If we just ignore
        # them for this evaluation, the cached values will simply pop
        # back once we turn on caching again.
        if not module.cached:
            for child in template.dependents(node):
                if cache.exists(fingerprints[child]):
                    print("clearing cached value for node %d: %s"
                          %(child, fingerprints[child]))
                    cache.delete(fingerprints[child])

        # Use cached value if it exists, skipping to the next loop iteration.
        if cache.exists(fingerprints[node]):
            print("retrieving cached value for node %d: %s"
                  %(node, fingerprints[node]))
            bundles = cache.retrieve(fingerprints[node])
            results.update((_key(node, k), v) for k, v in bundles.items())
            continue

        # Fields set for the current node
        template_fields = node_info.get('config', {})
        user_fields = config.get(str(node), {})

        # Evaluate the node
        print("calculating %s %s"%(node, module.id))
        outputs = _eval_node(node_id, module, inputs, template_fields, user_fields)

        # Collect the outputs
        bundles = {}
        for terminal in module.outputs:
            tid = terminal["id"]
            bundles[tid] = _bundle(terminal, outputs[tid])
        #print "caching", module.id, bundles
        #print "caching",_serialize(bundles, module.outputs)
        if module.cached:
            print("caching %s %s %s"%(node, module.id, fingerprints[node]))
            cache.store(fingerprints[node], bundles)
        results.update((_key(node, k), v) for k, v in bundles.items())

    #print list(sorted(results.keys()))

    if return_node is None:
        return results
    else:
        return results[_key(return_node, return_terminal)]

def _bundle(terminal, values):
    """
    Build a bundle for the terminal values.  The bundle has to carry the
    datatype around so that it has enough information to convert values
    to and from JSON.
    """
    datatype = lookup_datatype(terminal["datatype"])
    return Bundle(datatype=datatype, values=values)

def _key(node, terminal):
    """
    Combine *node* and *terminal* to a key into the results set for the
    current template.

    Returns a string unique within the template that can be used in JSON.
    """
    return ":".join((str(node), str(terminal)))


def _get_inputs(results, input_wires, input_terminals):
    """
    Lookup the inputs to all terminals from the results dictionary.

    *results* are the previous results.

    *input_wires* are all wires connecting to the input terminals for this
    node.  Each wire is represented by a dictionary linking the upstream
    source to the nodes target terminal as
    *{source: [node, terminal], target: [node, terminal]}*

    *input_terminals* is the list of input terminals for the node.

    Returns *{terminal: [dataset, ...]}* for all input terminals.
    """
    # Create bundles for the input terminals
    inputs = dict((terminal["id"], []) for terminal in input_terminals)
    for wire in input_wires:
        source_node, source_terminal = wire["source"]
        target_node, target_terminal = wire["target"]
        # Make key a tuple; wire["source"] won't work because it is
        # a mutable list, and can't be used as a dictionary key, but
        # an (int, string) tuple can be.
        v = results[_key(source_node, source_terminal)].values
        inputs[target_terminal].extend(v)
    return inputs


def _eval_node(node_id, module, inputs, template_fields, user_fields):
    """
    Run the action for the node.

    *id* is a string identifier for the node, for errors and logging

    *module* is the module activated by the node.

    *inputs* contains the input terminal values as *{terminal: [data, ...]}*.

    *template_fields* contains the field values stored in the
    template as *{field: [value, ...]}*.

    *user_fields* contains the field values sent as part of the template
    config as *{field: [value, ...]}*.

    Returns the output terminal bundle as *(terminal: [data, ...]}*.
    """
    # If the first input terminal is a multiple input terminal, then the
    # action needs to be called once with the bundle.
    # If the input terminals are all single input, then the action
    # needs to be called once for each dataset in the bundle.
    # For mixed single/multiple inputs the first input must be single.
    # If the output is single, it is made into a length one bundle.
    # If the output is multiple, then it is assumed to already be
    # a bundle.  With mixed single input/multiple output, all outputs
    # are concatenated into one bundle.

    # Check arity of module. This is determined by the first input terminal.
    multiple = not module.inputs or module.inputs[0]["length"] == 0
    bundle_length = 1 if multiple else len(inputs[module.inputs[0]["id"]])

    # determine field values
    fields = dict((field["id"], field["default"]) for field in module.fields)
     # override with template
    fields.update((k, v) for k, v in template_fields.items() if k in fields)
    # override with config
    fields.update((k, v) for k, v in user_fields.items() if k in fields)

    # validate fields
    for par in module.fields:
        name = par["id"]
        values = fields[name]
        if not par['multiple']:
            values = [values] if values is not None else []
        values = [_validate_par(node_id, par, value) for value in values]
        if len(values) == 0:
            del fields[name]
        elif len(values) == bundle_length:
            fields[name] = values
        elif len(values) == 1:
            fields[name] = values * bundle_length
        else:
            raise ValueError("Need one value of %s for each dataset in %s"
                             % (name, node_id))
        #print "fields", node_id, name, values

    # validate input terminals
    for par in module.inputs:
        name = par["id"]
        values = inputs[name]
        #print("inputs", node_id, name, values)
        # for value in values: _check_datatype(par, value)
        if len(values) == 0:
            # If no inputs, then either send an empty list or None, depending
            # on whether the input terminal is expecting a list or a singleton.
            fields[name] = [(None if par["length"] == 1 else [])]*bundle_length
        elif par["length"] == 0:
            fields[name] = [values]*bundle_length
        elif len(values) == bundle_length:
            fields[name] = values
        elif len(values) == 1:
            fields[name] = values*bundle_length
        else:
            raise ValueError("Need one value of %s for each dataset"%name)

    # Allocate slots for results
    outputs = dict((terminal["id"], []) for terminal in module.outputs)

    for k in range(bundle_length):
        # set up inputs
        action_args = dict((name, values[k]) for name, values in fields.items())

        # perform action
        #print "args", node_id, k, action_args
        result = _do_action(module, **action_args)

        #print node_id,result
        # Gather outputs
        for terminal, data in zip(module.outputs, result):
            if terminal["length"] == 0:
                outputs[terminal["id"]].extend(data)
            else:
                outputs[terminal["id"]].append(data)

    return outputs


def _validate_par(node_id, par, value):
    """
    Check that the parameters have the right type and length.
    """
    try:
        #print "validating",par,value
        return validate(par, value)
    except ValueError as exc:
        annotate_exception(" in " + node_id, exc)
        raise


def _do_action(module, **action_args):
    """
    Perform the module action, returning the results as a list.

    Because we know the number of outputs expected (each one is a terminal),
    we can convert no outputs or a single output to lists of length 0 and 1
    respectively.  This makes the module actions more natural to write.
    """
    num_outputs = len(module.outputs)
    result = module.action(**action_args)
    if num_outputs == 1:
        result = [result]
    elif num_outputs == 0:
        result = []
    return result

def fingerprint_template(template, config):
    """
    run the fingerprint operation on the whole template, returning
    the dict of fingerprints (one per output terminal)
    """
    fingerprints = {}
    for node, inputs in template.ordered():
        # Get fingerprints for terminal inputs
        inputs_fp = []
        for wire in inputs:
            source_node, source_terminal = wire['source']
            target_node, target_terminal = wire['target']
            source_fp = fingerprints[source_node]
            inputs_fp.extend([source_terminal, target_terminal, source_fp])

        # Get field values
        node_config = config.get(str(node), {})

        # Get module id and version
        module = template.modules[node]

        # Create and store fingerprint
        fp = fingerprint_node(module, node_config, inputs_fp)
        fingerprints[node] = fp
        #print "template fp", node, module, node_config, inputs_fp, fp

    return fingerprints


def fingerprint_node(module, node_config, inputs_fp):
    """
    Create a unique sha1 hash for a module based on its attributes and inputs.
    """
    config = module.get('config', {}).copy()
    config.update(node_config)
    config_str = str(_format_ordered(config))
    current_module_version = lookup_module(module['module']).version
    parts = [module['module'], current_module_version, config_str] + inputs_fp
    return generate_fingerprint(parts)

def generate_fingerprint(parts):
    """
    Generate a fingerprint from string parts.
    """
    key = ":".join(parts)
    if IS_PY3:
        key = key.encode('utf-8')
    fp = hashlib.sha1(key).hexdigest()
    return fp

# new methods that keep everything ordered
def _format_ordered(value):
    if isinstance(value, dict):
        return list((k, _format_ordered(v)) for k, v in sorted(value.items()))
    elif isinstance(value, list):
        return [_format_ordered(v) for v in value]
    elif isinstance(value, tuple):
        return tuple(_format_ordered(v) for v in value)
    elif callable(value):
        print("fingerprinting function %s"%str(value))
        return getsource(value)
    elif hasattr(value, '__getstate__'):
        value_state = value.__getstate__()
        if value_state is None:
            return value
        else:
           print("fingerprinting class using getstate %s"%str(value))
           return [value.__class__.__name__, _format_ordered(value_state)]
    elif hasattr(value, '__dict__'):
        print("fingerprinting class using dict %s"%str(value))
        return [value.__class__.__name__, _format_ordered(value.__dict__)]
    else:
        return value


# ===== Test support ===
@contextlib.contextmanager
def push_seed(seed=None): # pragma no cover
    """
    Set a temporary seed to the numpy random number generator, restoring it
    at the end of the context.  If seed is None, then get a seed from
    /dev/urandom (or the windows analogue).
    """
    from numpy.random import get_state, set_state, seed as set_seed
    state = get_state()
    set_seed(seed)
    yield
    set_state(state)

def verify_examples(source_file, tests, target_dir, seed=1): # pragma no cover
    """
    Run a set of templates, comparing the results against previous results.

    *source_file* should be *__file__*, which is the name of the lookup_module
    containing the test templates.

    *tests* is a list of tuples *(filename, (template, config))* where
    *filename* is  the name of the file containing the expected results
    from running *(template, config)*.  For example::

        verify_examples(__file__, tests=[
            ('bt7_example.test', (bt7_template, bt7_config)),
            ])

    *target_dir* is the directory which contains the expected result from
    running the template with the given configuration.  If it is not given,
    then it defaults to the "tests" subdirectory of the directory containing
    *source_file*.

    *seed* is the numpy random number generator (RNG) seed to use for the test.
    Before running each test, the RNG state is captured, the seed is set,
    the test is run, and the state is restored.  This allows you to run an
    individual example with :func:`run_example` and get the same result.
    *seed* defaults to 1.

    For each test, the expected results should be found in *target_dir/filename*
    If this file doesn't exist then the actual results will be stored and the
    test will pass.  If the file does exist, then the expected results are
    loaded and compared to the actual results.  New test results should be
    added/updated in the instrument repository.

    Raises *AssertionError* if any tests fail, indicating which files contain
    the expected result and the actual result.
    """
    import tempfile
    import os
    import json
    from os.path import join, exists, dirname
    import logging; logging.basicConfig(level=logging.WARNING)

    # No cache, so default to memory cache

    errors = []
    for (filename, (template, config)) in tests:
        print("checking %s"%filename)
        with push_seed(seed):
            actual = process_template(template, config)
        target_path = join(target_dir, filename)
        actual_str = json.dumps(actual, sort_keys=True)
        if not exists(target_path):
            if not exists(dirname(target_path)):
                os.makedirs(dirname(target_path))
            with open(target_path, 'wb') as fid:
                fid.write(actual_str)
        else:
            with open(target_path, 'rb') as fid:
                target_str = fid.read()
            if not actual_str == target_str:
                actual_path = join(tempfile.gettempdir(), filename)
                if not exists(dirname(actual_path)):
                    os.makedirs(dirname(actual_path))
                with open(actual_path, 'wb') as fid:
                    fid.write(actual_str)
                errors.append("  %r does not match target %r"
                              % (actual_path, target_path))
    if errors: # pragma no cover
        raise AssertionError("\n".join(errors))

def run_example(template, config, seed=None, verbose=False): # pragma no cover
    import json

    if verbose:
        print('template: '+json.dumps(template, sort_keys=True, indent=2))
        print('config: '+json.dumps(config, sort_keys=True, indent=2))

    with push_seed(seed):
        result = process_template(template, config)

    if verbose:
        print('result: '+json.dumps(result, sort_keys=True, indent=2))
    for key, value in result.items():
        for output in value.get('output', []):
            if not isinstance(output, dict):
                #print key, 'plot: ', output.get_plottable()
                pass

# internal tests
def test_format_ordered():
    udict, odict = {'x': 2, 'a': 3}, [('a', 3), ('x', 2)]
    # Note: Leave the ufn function as a 1-liner.  The test relies on the format
    def ufn(a): return a
    class A(object):
        def __init__(self):
            self.x, self.a = 2, 3
    class A2:
        def __init__(self):
            self.x, self.a = 2, 3
    pairs = [
        (udict, odict),
        ({'first': udict, 'second': 'ple'}, [('first', odict), ('second', 'ple')]),
        ([1, udict, 3], [1, odict, 3]),
        ((1, udict, 3), (1, odict, 3)),
        (ufn, "    def ufn(a): return a\n"),
        (A(), ['A', odict]),
        (A2(), ['A2', odict]),
        ]

    for u, o in pairs:
        actual = _format_ordered(u)
        print("%s => %r =? %r"%(str(u), actual, o))
        assert actual == o



"""
1-D and 2-D rebinning code.
"""
import numpy as np

try:
    from typing import Sequence, Optional, Union
except ImportError:
    pass

def rebin(x, I, xo, dtype=np.float64):
    # type: (Sequence, Sequence, Sequence, Union[str, type]) -> np.ndarray
    """
    Rebin a vector.

    *x* are the existing bin edges.

    *xo* are the new bin edges.

    *I* are the existing counts (one fewer than edges).

    *dtype* is the type to use for the intensity vectors.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.
    """
    # Coerce to arrays and check shape
    I = np.asarray(I)
    x, xo = np.asarray(x, dtype='d'), np.asarray(xo, dtype='d')
    if len(x.shape) != 1 or len(xo.shape) != 1 or len(x)-1 != len(I):
        raise TypeError("input array incorrect shape %s"%str(I.shape))

    ix = _rebin_counts(x, I, xo, dtype=dtype)
    return ix


def _rebin_counts(x, I, xo, dtype):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray) -> None

    # Sort directions ascending vs. descending
    if x[0] > x[-1]:
        x, I = x[::-1], I[::-1]
    reverse_xo = xo[0] > xo[-1]
    if reverse_xo:
        xo = xo[::-1]

    cc = np.empty(len(x), dtype=np.float64)
    cc[0] = 0
    cc[1:] = np.cumsum(I)
    integral = np.interp(xo, x, cc)  # result is always float64
    ix = np.diff(integral)

    # Set proper return type, rounding to nearest integer for integral types.
    # This may add/remove counts if we hit exactly 0.5 on some bins.
    if dtype == np.float32:
        ix = np.asarray(ix, dtype=dtype)
    elif dtype != np.float64:
        ix = np.asarray(ix+0.5, dtype=dtype)

    return ix[::-1] if reverse_xo else ix


def rebin2d(x, y, I, xo, yo, dtype=np.float64):
    # type: (Sequence, Sequence, Sequence, Sequence, Sequence, Union[str, type]) -> np.ndarray
    """
    Rebin a matrix.

    x, y are the existing bin edges
    xo, yo are the new bin edges
    I is the existing counts (one fewer than edges in each direction)

    For example, with x representing the column edges in each row and
    y representing the row edges in each column, the following
    represents a uniform field::

        >>> from reductus.dataflow.lib.rebin import rebin2d
        >>> x, y = [0, 2, 4, 5], [0, 1, 3]
        >>> z = [[2, 2, 1], [4, 4, 2]]

    We can check this by rebinning with uniform size bins::

        >>> xo, yo = range(6), range(4)
        >>> rebin2d(y, x, z, yo, xo) # doctest: +SKIP
        array([[1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1.]])

    dtype is the type to use for the output intensity.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.  Default is the dtype
    of the input intensity.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.
    """
    # Coerce inputs to arrays
    I = np.asarray(I)
    x, y, xo, yo = (np.asarray(v, dtype='d') for v in (x, y, xo, yo))
    shape_in = (x.shape[0]-1, y.shape[0]-1)
    if shape_in != I.shape or any(len(v.shape) != 1 for v in (x, y, xo, yo)):
        raise TypeError("input array incorrect shape %s"%str(I.shape))

    Io = _rebin_counts_2D(x, y, I, xo, yo, dtype=dtype)
    return Io


def _rebin_counts_2D(x, y, I, xo, yo, dtype):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray) -> None

    # Arrange indices so they are ascending.
    # Slices are cheap---they don't make copies.
    if x[0] > x[-1]:
        x, I = x[::-1], I[::-1, :]
    if y[0] > y[-1]:
        y, I = y[::-1], I[:, ::-1]
    reverse_xo = xo[0] > xo[-1]
    if reverse_xo:
        xo = xo[::-1]
    reverse_yo = yo[0] > yo[-1]
    if reverse_yo:
        yo = yo[::-1]

    # Create a cumulative sum.  Use float64 for integers to minimize
    # overflow issues, even though it costs 2x in terms of space for int32.
    intermediate_dtype = np.float32 if dtype == np.float32 else np.float64
    csx = np.empty((len(x), len(y)-1), dtype=intermediate_dtype)
    csx[0, :] = 0
    csx[1:, :] = I
    np.cumsum(csx, out=csx, axis=0)

    xindices = np.interp(xo, x, np.arange(len(x)))  # result is float64
    xindices_whole = np.floor(xindices).astype(int)
    xindices_frac = np.fmod(xindices, 1.0)  # result is float64
    # The only way an index will match the highest bin edge is for lookups at
    # or outside the range of the source.  In this case the fractional portion
    # will always == 0.0, so clipping this to the bin edge below is safe and
    # allows us to use the source weights unmodified.
    xintegral = I[xindices_whole.clip(max=len(x)-2), :]  # result is I.dtype
    xintegral = np.asarray(xintegral, dtype=intermediate_dtype)  # int=>float
    xintegral *= xindices_frac[:, None]
    xintegral += csx[xindices_whole, :]
    del csx
    # rebinned over x
    ix = np.diff(xintegral, axis=0)
    # Note: inplace diff, which returns -delta in xintegral[0:-1], is only a
    # few percent faster, and is much more subject to implementation details
    # in the overlapping region, so don't use it.  Numpy 1.13 promises to
    # create a temporary for inplace operations on overlapping regions, so
    # the minor performance gains are likely to disappear anyway.
    #xintegral[:-1] -= xintegral[1:]
    #ix = xintegral[:-1]
    del xintegral

    csy = np.empty((len(xo)-1, len(y)), dtype=intermediate_dtype)
    csy[:, 0] = 0
    csy[:, 1:] = ix
    np.cumsum(csy, out=csy, axis=1)

    yindices = np.interp(yo, y, np.arange(len(y)))
    yindices_whole = np.floor(yindices).astype(int)
    yindices_frac = np.fmod(yindices, 1.0)
    yintegral = ix[:, yindices_whole.clip(max=len(y)-2)]
    # ix is already the correct dtype, so no need to change it
    yintegral *= yindices_frac[None, :]
    yintegral += csy[:, yindices_whole]
    del csy
    # rebinned over x and y
    ixy = np.diff(yintegral, axis=1)
    #yintegral[:-1] -= yintegral[1:]
    #ixy = yintegral[:-1]
    del yintegral

    if ixy.dtype != dtype:
        # If return type is float, then the intermediate type will be the
        # same float32/float64.  If the return type is integer then the
        # intermediate type will be float64 and needs to be converted to int.
        ixy = np.asarray(ixy+0.5, dtype=dtype)

    if reverse_xo:
        ixy = ixy[::-1, :]
    if reverse_yo:
        ixy = ixy[:, ::-1]
    return ixy


# ================ Test code ==================
def _check1d(from_bins, val, to_bins, target):
    """
    Test 1D rebinning *from_bins* => *to_bins* of counts *val* => *target*.

    The expected *target* must be precomputed in order to test that the right
    value is produced.

    Tests all combinations of increasing/decreasing source/target bin edges.

    Yields a sequence of test functions *t() -> None* which raise errors if
    the result is unexpected.  *t.description* is set to the name of the
    test variant so that nosetests will have slightly nicer error reporting.
    """
    target = np.asarray(target, dtype='d')
    for (f, F) in [(from_bins, val), (from_bins[::-1], val[::-1])]:
        for (t, T) in [(to_bins, target), (to_bins[::-1], target[::-1])]:
            test_name = "rebin %s->%s: %s->%s"%(f, t, F, T)
            def test():
                result = rebin(f, F, t)
                assert np.linalg.norm(T-result) < 1e-14, \
                    "rebin failed %s: %s "%(test_name, result)
            test.description = test_name
            yield test


def _test1d():
    # Split a value
    for t in _check1d([1, 2, 3, 4], [10, 20, 30],
                      [1, 2.5, 4], [20, 40]):
        t()

    # bin is a superset of rebin
    for t in _check1d([0, 1, 2, 3, 4], [5, 10, 20, 30],
                      [1, 2.5, 3], [20, 10]):
        t()

    # bin is a subset of rebin
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50],
                      [2.5, 3.5], [25]):
        t()

    # one bin to many
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50],
                      [2.1, 2.2, 2.3, 2.4], [2, 2, 2]):
        t()

    # many bins to one
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50],
                      [2.5, 4.5], [60]):
        t()


def _check2d(x, y, z, xo, yo, zo):
    # type: (Sequence, Sequence, Sequence, Sequence, Sequence, Sequence) -> None
    result = rebin2d(x, y, z, xo, yo)
    target = np.array(zo, dtype=result.dtype)
    assert result.dtype == target.dtype
    assert np.linalg.norm(target-result) < 1e-14, \
        "rebin2d failed for %s, %s->%s, %s\n%s\n%s" % (x, y, xo, yo, z, zo)


def _uniform_test(x, y):
    z = np.array([y], 'd') * np.array([x], 'd').T
    xedges = np.concatenate([(0, ), np.cumsum(x)])
    yedges = np.concatenate([(0, ), np.cumsum(y)])
    nx = int(np.round(xedges[-1]))
    ny = int(np.round(yedges[-1]))
    ox = np.arange(nx+1)
    oy = np.arange(ny+1)
    target = np.ones([nx, ny], 'd')
    return _check2d(xedges, yedges, z, ox, oy, target)


def _test2d():
    x, y, I = [0, 3, 5, 7], [0, 1, 3], [[3, 6], [2, 4], [2, 4]]
    xo, yo, Io = range(8), range(4), [[1]*3]*7
    x, y, I, xo, yo, Io = [np.array(A, 'd') for A in [x, y, I, xo, yo, Io]]

    # Try various types and orders on a non-square matrix
    _check2d(x, y, I, xo, yo, Io)
    _check2d(x[::-1], y, I[::-1, :], xo, yo, Io)
    _check2d(x, y[::-1], I[:, ::-1], xo, yo, Io)
    _check2d(x, y, I, [7, 3, 0], yo, [[4]*3, [3]*3])
    _check2d(x, y, I, xo, [3, 2, 0], [[1, 2]]*7)
    _check2d(y, x, I.T, yo, xo, Io.T)  # C vs. Fortran ordering

    # Test smallest possible result
    _check2d([-1, 2, 4], [0, 1, 3],
             [[3, 6], [2, 4]], [1, 2], [1, 2], [1])
    # subset/superset
    _check2d([0, 1, 2, 3], [0, 1, 2, 3],
             [[1]*3]*3, [0.5, 1.5, 2.5], [0.5, 1.5, 2.5], [[1]*2]*2)
    for dtype in ['uint8', 'uint16', 'uint32', 'uint64', 'float32', 'float64']:
        _check2d(
            [0, 1, 2, 3, 4], [0, 1, 2, 3, 4],
            np.array([[1]*4]*4, dtype=dtype),
            [-2, -1, 2, 5, 6], [-2, -1, 2, 5, 6],
            np.array([[0, 0, 0, 0], [0, 4, 4, 0], [0, 4, 4, 0], [0, 0, 0, 0]],
                      dtype=dtype)
            )
    # non-square test
    _uniform_test([1, 2.5, 4, 0.5], [3, 1, 2.5, 1, 3.5])
    _uniform_test([3, 2], [1, 2])


def test():
    _test1d()
    _test2d()


if __name__ == "__main__":
    test()



# -*- coding: latin-1 -*-
import numpy as np

def LoadUXDData(filename, path="", friendly_name=""):
    """ Load two-dimensional mesh scans from Bruker x-ray files """
    ###################
    # By André Guzmán #
    ###################

    #This program assumes that the step sizes of all Rocking Curves are the same
    #I do not know what will happen if they are not (I've never tried) - but the result will probably not be pretty

    ############# Variables #################################
    #Read from file header (applies to all ranges)
    wavelength = 0

    #Arrays
    data_array = [] #main array to return
    two_theta_array = [] #Isn't it good that I ended all these variables with _array - there's already a two_theta variable that it could've been mixed up with
    theta_array = [] #All of the tube angles for the various scans (Theta)
    counts_array = [] #Counts go here
    pixels_array = [] #This is all 1s
    monitor_array = [] #Does the *100 multiplication for detectorslit.in
    count_time_array = [] #Count times go here

    #Range variables
    #Range Constants (values that don't change within a range, but do change between ranges)
    curr_range = 0 #determines which range is being read
    prev_step_time = 0
    step_time = 0
    step_size = 0
    theta_start = 0
    two_theta = 0
    detector_slit = False
    #Range Counters (these change in the range - and most of them only count up)
    theta = 0
    range_counts_array = []
    range_theta_array = []
    ############# END Variables #################################

    file_obj = open(os.path.join(path, filename), 'r')
    np = numpy

    for lines in file_obj:
        line = lines.strip()

        if line != "" and line[0].isdigit():
            #add counts to the count array and other stuff
            range_theta_array.append(theta)
            theta += step_size
            range_counts_array.append(float(line))

        elif line.count("_WL1") != 0:
            nums = line.split(" = ")
            wavelength = float(nums[1].strip())
        elif line.count("; Data ") != 0:
            nums = line.split(" ")
            curr_range = int(nums[4].strip())
            #step_time = 0
            step_size = 0
            theta_start = 0
            two_theta = 0
            theta = 0

            if curr_range != 1:
                counts_array.append(range_counts_array)
                theta_array.append(range_theta_array)

                range_counts_array = []
                range_theta_array = []
        elif line.count("_STEPTIME") != 0:
            nums = line.split(" = ")
            step_time = float(nums[1].strip()) - prev_step_time #Step time increments for some reason, this fixes that
            prev_step_time = float(nums[1].strip())
            count_time_array.append(step_time)
        elif line.count("_STEPSIZE") != 0:
            nums = line.split(" = ")
            step_size = float(nums[1].strip())
        elif line.count("_START") != 0:
            nums = line.split(" = ")
            theta_start = float(nums[1].strip())
            theta = theta_start
        elif line.count("_2THETA") != 0:
            nums = line.split(" = ")
            two_theta = float(nums[1].strip())
            two_theta_array.append(two_theta)
        elif line.count("_DETECTORSLIT") != 0:
            nums = line.split(" = ")
            slit = nums[1].strip()

            if slit == "in":
                detector_slit = True
                monitor_array.append(1)
            else:
                detector_slit = False
                monitor_array.append(0.01)
        else:
            two = 1 + 1
            #This really doesn't need to be here (at all), but I like IF statements that end in ELSEs

    #For Loop skips over the last range (which might be important), so its added here
    if curr_range != 1:
        # Adding arrays to other arrays and resetting them
        counts_array.append(range_counts_array)
        theta_array.append(range_theta_array)

        range_counts_array = []
        range_theta_array = []

    #Now that we're done with reading the file, its time to set up the MetaArray
    file_obj.close() # We're done with the file now

    for i in range(curr_range):
        pixels_array.append(1)

    theta_range = theta_array[-1][-1] - theta_array[0][0]
    theta_elements = int(theta_range / step_size + 0.5) + 1 # Assumes the step sizes of all the ranges are the same

    data_array = np.zeros((theta_elements, curr_range, 4))
    data_theta_array = [] #Rhyming was totally intentional

    theta_incr = theta_array[0][0] # This is a bad name for this variable
    for i in range(theta_elements):
        data_theta_array.append(theta_incr)
        theta_incr += step_size

    for i in range(len(counts_array)):
        start = int((theta_array[i][0] - data_theta_array[0])/step_size + 0.5) #Add 0.5 for rounding - the int() function does strange things to floats
        stop = int(int(start + ((theta_array[i][-1] - theta_array[i][0])/step_size) + .5) + 1) #Add 1 because arrays that go from 0 to N have N+1 elements

        data_array[start:stop, i, 0] = counts_array[i]
        data_array[start:stop, i, 1] = pixels_array[i]
        data_array[start:stop, i, 2] = monitor_array[i]
        data_array[start:stop, i, 3] = count_time_array[i]

    info = [
        {"name": "theta", "units": "degrees", "values": data_theta_array},
        {"name": "twotheta", "units": "degrees", "values": two_theta_array}]
    info.append({"name": "Measurements", "cols": [
                        {"name": "counts"},
                        {"name": "pixels"},
                        {"name": "monitor"},
                        {"name": "count_time"}]})
    info.append({"filename": filename, "friendly_name": friendly_name, "path":path, "CreationStory": ""})

    data = MetaArray(data_array, dtype='float', info=info)
    return data

def LoadUXDMany(filedescriptors):
    result = []
    for fd in filedescriptors:
        new_data = LoadUXDData(fd['filename'], friendly_name=fd['friendly_name'])
        if isinstance(new_data, list):
            result.extend(new_data)
        else:
            result.append(new_data)
    return result



from posixpath import basename, join
from copy import copy, deepcopy
from io import BytesIO
import numpy as np

ALL_ACTIONS = []

def cache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = True
    return action

def nocache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = False
    return action

def module(action):
    """
    Decorator which records the action in *ALL_ACTIONS*.

    This just collects the action, it does not otherwise modify it.
    """
    ALL_ACTIONS.append(action)

    # This is a decorator, so return the original function
    return action

#################
# Loader stuff
#################

def findPeak(qvals, counts, min_height=10, sample_width=10):
    # using first moment to get center and height
    # IGOR command: FindPeak/P/Q/M=10/R=[(temp-10),(temp+10)]
    # where 
    # /M=minLevel Defines minimum level of a peak 
    # /R=[startP,endP] Specifies point range and direction for search
    peak_index = np.argmax(counts)
    peak_q = qvals[peak_index]
    peak_val = counts[peak_index]
    if peak_val < min_height:
        return None,None
    
    central_slice = slice(peak_index-sample_width, peak_index+sample_width+1)
    central_counts = counts[central_slice]
    central_qvals = qvals[central_slice]
    peak_center = np.sum(central_qvals*central_counts)/np.sum(central_counts) 
    return {"peak_center": peak_center, "peak_value": peak_val}

@cache
@module
def LoadRawUSANS(filelist=None, check_timestamps=True, det_deadtime=7e-6, trans_deadtime=1.26e-5):
    """
    loads a data file into a RawSansData obj and returns that.

    **Inputs**

    filelist (fileinfo[]): Files to open.
    
    check_timestamps (bool): verify that timestamps on file match request

    det_deadtime {main deadtime (s)} (float): main detector deadtime, in seconds

    trans_deadtime {trans deadtime (s)} (float): transmission detector deadtime, in seconds 

    **Returns**

    output (data[]): all the entries loaded.

    2020-01-29 Brian Maranville
    """
    from reductus.dataflow.fetch import url_get
    from .loader import readUSANSNexus
    from .usansdata import USansData
    if filelist is None:
        filelist = []
    data = []
    for fileinfo in filelist:
        path, mtime, entries = fileinfo['path'], fileinfo.get('mtime', None), fileinfo.get('entries', None)
        name = basename(path)
        fid = BytesIO(url_get(fileinfo, mtime_check=check_timestamps))
        entries = readUSANSNexus(name, fid, det_deadtime=det_deadtime, trans_deadtime=trans_deadtime)
        
        data.extend(entries)

    return data

@cache
@module
def convert_to_countrate(unnormalized, do_mon_norm=True, mon0=1e6):
    """"
    Given a USansData object, normalize the data to time and the provided monitor

    **Inputs**

    unnormalized (data): data in

    do_mon_norm {normalize to mon0} (bool) : normalized detCnt to specified mon0 (otherwise just time)

    mon0 (float): provided monitor

    **Returns**

    output (data): corrected for dead time

    2020-01-28 Brian Maranville
    """

    time = unnormalized.countTime
    monitor = unnormalized.monCts

    if do_mon_norm:
        unnormalized.detCts *= mon0/monitor
    else:
        unnormalized.detCts /= time
    
    unnormalized.monCts /= time
    unnormalized.transCts /= time
        
    return unnormalized

def _findTWideCts(data, q_threshold=1e-4):
    mask = (data.Q > q_threshold)
    missing_data_warn = False

    if mask.sum() == 0:
        mask = np.zeros_like(data.Q, dtype='bool')
        mask[-4:] = True
        missing_data_warn = True

    mean_cts = data.transCts[mask].mean()
    mean_mon = data.monCts[mask].mean()

    twide = mean_cts/mean_mon

    return twide, missing_data_warn

@module
def findTWideCts(data, q_threshold=1e-4):
    """"
    Given a USansData object, find the TWide parameters.
    If no data is available above the Q threshold, uses the last 4 points and returns
    a warning in the parameters.

    **Inputs**

    data (data): data in

    q_threshold (float): use Q values larger than this to do the calculation

    **Returns**

    twide (params): corrected for dead time

    2020-01-28 Brian Maranville
    """
    from reductus.sansred.sansdata import Parameters

    twide, missing_data_warn = _findTWideCts(data, q_threshold)

    output = {"fileNumber": data.metadata["run.instFileNum"], "TWIDE": twide.x, "TWIDE_ERR": np.sqrt(twide.variance)}
    if missing_data_warn:
        output["WARNING"] = "You don't have data past 1e-4 A-1 (~2 degrees) - so Twide may not be reliable"
    
    return Parameters(params=output)

@cache
@module
def setPeakCenter(data, peak_params, peak_center=None):
    """"
    Given a USansData object and a new peak center, adjusts Q values to match the new center
    If a peak_center is specified directly, it will always be used
    Otherwise the value from the input peak_params (coming from the fit) will be used
    
    **Inputs**

    data (data): data in

    peak_params (params?): send output from findPeak to here, and the 'peak_center' will be used unless overridden

    peak_center (float*): override the value from the peak_params['peak_center']

    **Returns**

    adjusted_data (data): corrected for new peak center

    2020-01-28 Brian Maranville
    """
    peak_center = peak_center if peak_center is not None else peak_params.params['peak_center']

    data.Q -= peak_center
    data.Q_offset = peak_center

    return data

@cache
@module
def getPeakParams(data):
    """"
    Given a USansData object peak parameters (center and height) will be calculated
    
    **Inputs**

    data (data): data in

    **Returns**

    peak_params (params): peak data

    2020-01-29 Brian Maranville
    """
    from reductus.sansred.sansdata import Parameters

    peak_params = {"fileNumber": data.metadata["run.instFileNum"], "intent": data.metadata["analysis.intent"]}
    peak_params.update(findPeak(data.Q, data.detCts.x))

    return Parameters(params=peak_params)

@cache
@module
def correctData(sample, empty, bkg_level=0.0, emp_level=0.0, thick=1.0, dOmega=7.1e-7):
    """"
    Do the final data reduction.  Requires a data and empty, normalized.
    
    **Inputs**

    sample (data): data in

    empty (data): empty in

    bkg_level (float): background level

    emp_level (float): empty background level

    thick (float): thickness of sample, in cm

    dOmega (float): solid angle of detector (steradians)

    **Returns**

    corrected (data): corrected output

    corrected_info (params): correction info

    2020-01-29 Brian Maranville
    """
    from reductus.dataflow.lib.uncertainty import Uncertainty
    from .usansdata import USansCorData
    from reductus.sansred.sansdata import Parameters
    # find q-range of empty:
    empty_qmax = empty.Q.max()
    empty_qmin = empty.Q.min()
    empty_mask = np.logical_and(sample.Q >= empty_qmin, sample.Q <= empty_qmax)
    interpolated_empty = np.interp(sample.Q, empty.Q, empty.detCts.x, left=emp_level, right=emp_level)
    interpolated_empty_err = np.interp(sample.Q, empty.Q, empty.detCts.variance, left=0, right=0)
    tempI = Uncertainty(interpolated_empty, interpolated_empty_err)

    Twide_sam, _ = _findTWideCts(sample)
    Twide_emp, _ = _findTWideCts(empty)

    pkHtEMP = empty.detCts.x.max()
    pkHtSAM = sample.detCts.x.max()

    Trock = pkHtSAM / pkHtEMP
    Twide = Twide_sam / Twide_emp

    ratio = Trock / Twide

    iqCOR = sample.detCts - Trock*tempI - (1-Trock)*bkg_level

    scale = 1/(Twide*thick*dOmega*pkHtEMP)

    iqCOR *= scale

    info = {
        "Sample file": sample.metadata["run.filename"],
        "Empty file": empty.metadata["run.filename"],
        "Trock/Twide": ratio.x,
        "Thickness": thick,
        "Twide": Twide.x,
        "Trock": Trock,
        "Sample Peak Angle": getattr(sample, 'Q_offset', 0.0),
        "Empty Peak Angle": getattr(empty, 'Q_offset', 0.0),
        "Empty level": emp_level,
        "Bkg level": bkg_level,
        "dQv": sample.metadata["dQv"],
        "Start time": sample.metadata["start_time"],
        "Entry": sample.metadata["entry"],
    }

    corrected = USansCorData(metadata=info,iqCOR=iqCOR, Q=sample.Q)
    
    return corrected, Parameters(info)

@cache
@module
def correctJoinData(sample, empty, q_tol=0.01, bkg_level=0.0, emp_level=0.0, thick=1.0, dOmega=7.1e-7):
    """"
    Do the final data reduction.  Requires sample and empty datasets, normalized.
    Multiple inputs in either will be joined into one dataset before processing.
    
    **Inputs**

    sample (data[]): data in

    empty (data): empty in

    q_tol (float): values closer together than this tolerance in Q will be joined into a single point

    bkg_level (float): background level

    emp_level (float): empty background level

    thick (float): thickness of sample, in cm

    dOmega (float): solid angle of detector (steradians)

    **Returns**

    corrected (cor): corrected output

    corrected_info (params[]): correction info

    2020-01-29 Brian Maranville
    """

    from reductus.dataflow.lib.uncertainty import Uncertainty
    from .usansdata import USansCorData
    from reductus.sansred.sansdata import Parameters


    data_groups = make_groups([s.Q for s in sample])

    reduced_pairs = [correctData(s, empty, bkg_level=bkg_level, emp_level=emp_level, thick=thick, dOmega=dOmega) for s in sample]

    reduced_values = [r[0] for r in reduced_pairs]
    reduced_infos = [r[1].params for r in reduced_pairs]

    Qvals = []
    iqCOR_x = []
    iqCOR_variance = []
    for g in data_groups:
        qmean = np.array([x for x,i,j in g]).mean()
        iq = [reduced_values[i].iqCOR[j] for x,i,j in g]
        iqmean = sum(iq) / len(iq)
        Qvals.append(qmean)
        iqCOR_x.append(iqmean.x)
        iqCOR_variance.append(iqmean.variance)
    
    iqCOR = Uncertainty(iqCOR_x, iqCOR_variance)
    Qvals = np.array(Qvals)

    corrected = USansCorData(metadata=reduced_infos,iqCOR=iqCOR, Q=Qvals)
    
    return corrected, [Parameters(info) for info in reduced_infos]

@module
def crop_corrected(data, start_stop=[None, None]):
    """
    Crop USANS corrected data

    **Inputs**

    data (cor): USANS corrected input

    start_stop {Crop start and stop} (range?:x) : Select which part of the data (along x-axis) to keep

    **Returns**

    cropped (cor): cropped data

    | 2020-05-04 Brian Maranville
    """

    if start_stop is None:
        start_stop = [None, None]
    
    cropped = copy(data)
    start, stop = start_stop
    left_index = None if start is None else np.searchsorted(data.Q, start, side='left')
    right_index = None if stop is None else np.searchsorted(data.Q, stop, side='right')
    new_slice = slice(left_index, right_index)
    cropped.Q = data.Q[new_slice]
    cropped.iqCOR = data.iqCOR[new_slice]
    return cropped
    

def make_groups(xvals_list, xtol=0):
    # xvals_list is list of xvals arrays
    groups = []
    for i, xvals in enumerate(xvals_list):
        for j, x in enumerate(xvals):
            matched = False
            for g in groups:
                d = [(x-gp[0]) <= (xtol * x) for gp in g]
                if all([(x-gp[0]) <= (xtol * x) for gp in g]):
                    g.append((x,i,j))
                    matched = True
                    break
            if not matched:
                groups.append([(x,i,j)])
    return groups


            




r"""
Convert a restructured text document to html.

Inline math markup can uses the *math* directive, or it can use latex
style *\$expression\$*.  Math is rendered using simple html and unicode,
not mathjax.
"""
from __future__ import print_function

import sys
import re
from contextlib import contextmanager

from docutils.core import publish_parts
from docutils.writers.html4css1 import HTMLTranslator
from docutils.nodes import SkipNode

# As of 2020-03-01 latest mathjax is 2.7.7 and 3.0.1
MATHJAX_CONFIG = "mathjax https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/MathJax.js"

def rst2html(rst, part="whole", math_output="html"):
    r"""
    Convert restructured text into simple html.

    Valid *math_output* formats for formulas include:
    - html
    - mathml
    - mathjax
    See the `docutils documentation <http://docutils.sourceforge.net/docs/user/config.html#math-output>`_
    for details.

    The following *part* choices are available:
    - whole: the entire html document
    - html_body: document division with title and contents and footer
    - body: contents only

    There are other parts, but they don't make sense alone:

        subtitle, version, encoding, html_prolog, header, meta,
        html_title, title, stylesheet, html_subtitle, html_body,
        body, head, body_suffix, fragment, docinfo, html_head,
        head_prefix, body_prefix, footer, body_pre_docinfo, whole
    """
    # Ick! mathjax doesn't work properly with math-output, and the
    # others don't work properly with math_output!
    if math_output == "mathjax":
        settings = {"math_output": MATHJAX_CONFIG}
    else:
        settings = {"math-output": math_output}

    # math2html and mathml do not support \frac12
    rst = replace_compact_fraction(rst)

    # mathml, html do not support \tfrac
    if math_output in ("mathml", "html"):
        rst = rst.replace(r'\tfrac', r'\frac')

    rst = replace_dollar(rst)
    with suppress_html_errors():
        parts = publish_parts(source=rst, writer_name='html',
                              settings_overrides=settings)
    return parts[part]

@contextmanager
def suppress_html_errors():
    r"""
    Context manager for keeping error reports out of the generated HTML.

    Within the context, system message nodes in the docutils parse tree
    will be ignored.  After the context, the usual behaviour will be restored.
    """
    visit_system_message = HTMLTranslator.visit_system_message
    HTMLTranslator.visit_system_message = _skip_node
    yield None
    HTMLTranslator.visit_system_message = visit_system_message

def _skip_node(self, node):
    try:
        doc = node.document.nameids.keys()[0]
        print("error while processing line %s of %s rst" % (node.line+1, doc),
              file=sys.stderr)
    except Exception:
        pass
    raise SkipNode


_compact_fraction = re.compile(r"(\\[cdt]?frac)([0-9])([0-9])")
def replace_compact_fraction(content):
    r"""
    Convert \frac12 to \frac{1}{2} for broken latex parsers
    """
    return _compact_fraction.sub(r"\1{\2}{\3}", content)


_dollar = re.compile(r"(?:^|(?<=\s|[(]))[$]([^\n]*?)(?<![\\])[$](?:$|(?=\s|[.,;)\\]))")
_notdollar = re.compile(r"\\[$]")
def replace_dollar(content):
    r"""
    Convert dollar signs to inline math markup in rst.
    """
    content = _dollar.sub(r":math:`\1`", content)
    content = _notdollar.sub("$", content)
    return content


def test_dollar():
    assert replace_dollar(u"no dollar") == u"no dollar"
    assert replace_dollar(u"$only$") == u":math:`only`"
    assert replace_dollar(u"$first$ is good") == u":math:`first` is good"
    assert replace_dollar(u"so is $last$") == u"so is :math:`last`"
    assert replace_dollar(u"and $mid$ too") == u"and :math:`mid` too"
    assert replace_dollar(u"$first$, $mid$, $last$") == u":math:`first`, :math:`mid`, :math:`last`"
    assert replace_dollar(u"dollar\\$ escape") == u"dollar$ escape"
    assert replace_dollar(u"dollar \\$escape\\$ too") == u"dollar $escape$ too"
    assert replace_dollar(u"spaces $in the$ math") == u"spaces :math:`in the` math"
    assert replace_dollar(u"emb\\ $ed$\\ ed") == u"emb\\ :math:`ed`\\ ed"
    assert replace_dollar(u"$first$a") == u"$first$a"
    assert replace_dollar(u"a$last$") == u"a$last$"
    assert replace_dollar(u"$37") == u"$37"
    assert replace_dollar(u"($37)") == u"($37)"
    assert replace_dollar(u"$37 - $43") == u"$37 - $43"
    assert replace_dollar(u"($37, $38)") == u"($37, $38)"
    assert replace_dollar(u"a $mid$dle a") == u"a $mid$dle a"
    assert replace_dollar(u"a ($in parens$) a") == u"a (:math:`in parens`) a"
    assert replace_dollar(u"a (again $in parens$) a") == u"a (again :math:`in parens`) a"

if __name__ == "__main__":
    test_dollar()



r"""
Weighted linear and polynomial solver with uncertainty.

Given $A \bar x = \bar y \pm \delta \bar y$, solve using *s = wsolve(A,y,dy)*

*wsolve* uses the singular value decomposition for increased accuracy.

The uncertainty in the solution is estimated from the scatter in the data.
Estimates the uncertainty for the solution from the scatter in the data.

The returned model object *s* provides:

    ======== ============================================
    ======== ============================================
    s.x      solution
    s.std    uncertainty estimate assuming no correlation
    s.rnorm  residual norm
    s.DoF    degrees of freedom
    s.cov    covariance matrix
    s.ci(p)  confidence intervals at point p
    s.pi(p)  prediction intervals at point p
    s(p)     predicted value at point p
    ======== ============================================

**Example**

Weighted system::

    >>> import numpy as np
    >>> A = np.array([[1,2,3],[2,1,3],[1,1,1]], 'd')
    >>> dy = [0.2, 0.01, 0.1]
    >>> y = [ 14.16, 13.01, 6.15]
    >>> s = wsolve(A, y, dy)
    >>> print(", ".join("%0.2f +/- %0.2f"%(a, b) for a, b in zip(s.x.flatten(), s.std)))
    1.05 +/- 0.17, 2.20 +/- 0.12, 2.91 +/- 0.12


Note there is a counter-intuitive result that scaling the estimated
uncertainty in the data does not affect the computed uncertainty in
the fit.  This is the correct result --- if the data were indeed
selected from a process with ten times the uncertainty, you would
expect the scatter in the data to increase by a factor of ten as
well.  When this new data set is fitted, it will show a computed
uncertainty increased by the same factor.  Monte carlo simulations
bear this out.  The conclusion is that the dataset carries its own
information about the variance in the data, and the weight vector
serves only to provide relative weighting between the points.
"""

__all__ = ['wsolve', 'wpolyfit', 'LinearModel', 'PolynomialModel', 'smooth']

# FIXME: test second example
#
# Example 2: weighted overdetermined system  y = x1 + 2*x2 + 3*x3 + e
#
#    A = fullfact([3, 3, 3]); xin=[1;2;3];
#    y = A*xin; dy = rand(size(y))/50; y+=dy.*randn(size(y));
#    [x, s] = wsolve(A, y, dy);
#    dx = s.normr*sqrt(sumsq(inv(s.R'))'/s.df);
#    res = [xin, x, dx]


import numpy as np


class LinearModel(object):
    r"""
    Model evaluator for linear solution to $Ax = y$.

    Use *s(A)* to compute the predicted value of the linear model *s*
    at points given on the rows of $A$.

    Computes a confidence interval (range of likely values for the
    mean at $x$) or a prediction interval (range of likely values
    seen when measuring at $x$).  The prediction interval gives
    the width of the distribution at $x$.  This should be the same
    regardless of the number of measurements you have for the value
    at $x$.  The confidence interval gives the uncertainty in the
    mean at $x$.  It should get smaller as you increase the number of
    measurements.  Error bars in the physical sciences usually show
    a $1-\alpha$ confidence value of $\text{erfc}(1/\sqrt{2})$, representing
    a $1-\sigma$ standand deviation of uncertainty in the mean.

    Confidence intervals for the expected value of the linear system
    evaluated at a new point $w$ are given by the $t$ distribution for
    the selected interval $1-\alpha$, the solution $x$, and the number
    of degrees of freedom $n-p$:

    .. math::

        w^T x \pm t^{\alpha/2}_{n-p} \sqrt{ \text{var}(w) }

    where the variance $\text{var}(w)$ is given by:

    .. math::

        \text{var}(w) = \sigma^2 (w^T (A^TA)^{-1} w)

    Prediction intervals are similar, except the variance term increases to
    include both the uncertainty in the predicted value and the variance in
    the data:

    .. math::

        \text{var}(w) = \sigma^2 (1 + w^T (A^TA)^{-1} w)
    """
    def __init__(self, x=None, DoF=None, SVinv=None, rnorm=None):
        # Note: SVinv should be computed from S,V where USV' = A
        #: solution to the equation $Ax = y$
        self.x = x
        #: number of degrees of freedom in the solution
        self.DoF = DoF
        #: 2-norm of the residuals $||y-Ax||_2$
        self.rnorm = rnorm
        self._SVinv = SVinv

    def __call__(self, A):
        """
        Return the prediction for a linear system at points in the rows of A.
        """
        return np.dot(np.asarray(A), self.x)

    # covariance matrix invC = A'A  = (USV')'USV' = VSU'USV' = VSSV'
    # C = inv(A'A) = inv(VSSV') = inv(V')inv(SS)inv(V) = Vinv(SS)V'
    # diag(inv(A'A)) is sum of the squares of the columns inv(S) V'
    # and is also the sum of the squares of the rows of V inv(S)
    @property
    def cov(self):
        """covariance matrix [inv(A'A); O(n^3)]"""
        # FIXME: don't know if we need to scale by C, but it will
        # at least make things consistent
        C = self.rnorm**2/self.DoF if self.DoF > 0 else 1
        return C * np.dot(self._SVinv, self._SVinv.T)

    @property
    def var(self):
        """solution variance [diag(cov); O(n^2)]"""
        C = self.rnorm**2/self.DoF if self.DoF > 0 else 1
        return C * np.sum(self._SVinv**2, axis=1)

    @property
    def std(self):
        """solution standard deviation [sqrt(var); O(n^2)]"""
        return np.sqrt(self.var)

    @property
    def p(self):
        """p-value probability of rejection"""
        from scipy.stats import chi2  # lazy import in case scipy not present
        return chi2.sf(self.rnorm ** 2, self.DoF)

    def _interval(self, X, alpha, pred):
        """
        Helper for computing prediction/confidence intervals.
        """
        # Comments from QR decomposition solution to Ax = y:
        #
        #   Rather than A'A we have R from the QR decomposition of A, but
        #   R'R equals A'A.  Note that R is not upper triangular since we
        #   have already multiplied it by the permutation matrix, but it
        #   is invertible.  Rather than forming the product R'R which is
        #   ill-conditioned, we can rewrite x' inv(A'A) x as the equivalent
        #      x' inv(R) inv(R') x = t t', for t = x' inv(R)
        #
        # We have since switched to an SVD solver, which gives us
        #
        #    invC = A' A  = (USV')' USV' = VSU' USV' = V S S V'
        #    C = inv(A'A) = inv(VSSV') = inv(V') inv(S S) inv(V)
        #      = V inv(S S) V' = V inv(S) inv(S) V'
        #
        # Substituting, we get
        #
        #    x' inv(A'A) x = t t', for t = x' V inv(S)
        #
        # Since x is a vector, t t' is the inner product sum(t**2).
        # Note that LAPACK allows us to do this simultaneously for many
        # different x using sqrt(sum(T**2,axis=1)), with T = X' Vinv(S).
        #
        # Note: sqrt(F(1-a;1,df)) = T(1-a/2;df)
        #
        from scipy.stats import t  # lazy import in case scipy not present
        y = np.dot(X, self.x).ravel()
        s = t.ppf(1-alpha/2, self.DoF) * self.rnorm/np.sqrt(self.DoF)
        t = np.dot(X, self._SVinv)
        dy = s * np.sqrt(pred + np.sum(t**2, axis=1))
        return y, dy

    def ci(self, A, sigma=1):
        r"""
        Compute the calculated values and the confidence intervals
        for the linear model evaluated at $A$.

        *sigma=1* corresponds to a $1-\sigma$ confidence interval

        Confidence intervals are sometimes expressed as $1-\alpha$ values,
        where $\alpha = \text{erfc}(\sigma/\sqrt{2})$.
        """
        from scipy.special import erfc  # lazy import in case scipy not present
        alpha = erfc(sigma / np.sqrt(2))
        return self._interval(np.asarray(A), alpha, 0)

    def pi(self, A, p=0.05):
        r"""
        Compute the calculated values and the prediction intervals
        for the linear model evaluated at $A$.

        *p=0.05* corresponds to the 95% prediction interval.
        """
        return self._interval(np.asarray(A), p, 1)


def wsolve(A, y, dy=1, rcond=1e-12):
    r"""
    Given a linear system $y = A x + \delta y$, estimates $x$ and $\delta x$.

    *A* is an n x m array of measurement points.

    *y* is an n x k array or vector of length n of measured values at *A*.

    *dy* is a scalar or an n x 1 array of uncertainties in the values at *A*.

    Returns :class:`LinearModel`.
    """
    # The ugliness v[:, N.newaxis] transposes a vector
    # The ugliness N.dot(a, b) is a*b for a,b matrices
    # The ugliness vh.T.conj() is the hermitian transpose

    # Make sure inputs are arrays
    A, y, dy = np.asarray(A), np.asarray(y), np.asarray(dy)
    if dy.ndim == 1:
        dy = dy[:, np.newaxis]
    if y.ndim == 1:
        y = y[:, np.newaxis]

    # Apply weighting if dy is not a scalar
    # If dy is a scalar, it cancels out of both sides of the equation
    # Note: with A,dy arrays instead of matrices, A/dy operates element-wise
    # Since dy is a row vector, this divides each row of A by the corresponding
    # element of dy.
    if dy.ndim == 2:
        A, y = A/dy, y/dy

    # Singular value decomposition: A = U S V.H
    # Since A is an array, U, S, VH are also arrays
    # The zero indicates an economy decomposition, with u nxm rathern than nxn
    u, s, vh = np.linalg.svd(A, 0)

    # FIXME what to do with ill-conditioned systems?
    # Use regularization? L1 (Lasso), L2 (Ridge) or both (Elastic Net)
    #if s[-1]<rcond*s[0]: raise ValueError, "matrix is singular"
    # s[s<rcond*s[0]] = 0.  # Can't do this because 1/s below will fail

    # Solve: x = V inv(S) U.H y
    # S diagonal elements => 1/S is inv(S)
    # A*D, D diagonal multiplies each column of A by the corresponding diagonal
    # D*A, D diagonal multiplies each row of A by the corresponding diagonal
    # Computing V*inv(S) is slightly faster than inv(S)*U.H since V is smaller
    # than U.H.  Similarly, U.H*y is somewhat faster than V*U.H
    SVinv = vh.T.conj() / s
    Uy = np.dot(u.T.conj(), y)
    x = np.dot(SVinv, Uy)

    DoF = y.shape[0] - x.shape[0]
    rnorm = np.linalg.norm(y - np.dot(A, x))

    return LinearModel(x=x, DoF=DoF, SVinv=SVinv, rnorm=rnorm)


def _poly_matrix(x, degree, origin=False):
    """
    Generate the matrix A used to fit a polynomial using a linear solver.
    """
    if origin:
        n = np.array(range(degree, 0, -1))
    else:
        n = np.array(range(degree, -1, -1))
    return np.asarray(x)[:, None] ** n[None, :]


class PolynomialModel(object):
    r"""
    Model evaluator for best fit polynomial $p(x) = y +/- \delta y$.

    Use *p(x)* for PolynomialModel *p* to evaluate the polynomial at all
    points in the vector *x*.
    """

    def __init__(self, x, y, dy, s, origin=False):
        self.x, self.y, self.dy = [np.asarray(v) for v in (x, y, dy)]
        #: True if polynomial goes through the origin
        self.origin = origin
        #: polynomial coefficients
        self.coeff = np.ravel(s.x)
        if origin:
            self.coeff = np.hstack((self.coeff, 0))
        #: polynomial degree
        self.degree = len(self.coeff) - 1
        #: number of degrees of freedom in the solution
        self.DoF = s.DoF
        #: 2-norm of the residuals $||y-Ax||_2$
        self.rnorm = s.rnorm
        self._conf = s

    @property
    def cov(self):
        """
        covariance matrix

        Note that the ones column will be absent if *origin* is True.
        """
        return self._conf.cov

    @property
    def var(self):
        """solution variance"""
        return self._conf.var

    @property
    def std(self):
        """solution standard deviation"""
        return self._conf.std

    @property
    def p(self):
        """p-value probability of rejection"""
        return self._conf.p

    def __call__(self, x):
        """
        Evaluate the polynomial at x.
        """
        return np.polyval(self.coeff, x)

    def der(self, x):
        """
        Evaluate the polynomial derivative at x.
        """
        return np.polyval(np.polyder(self.coeff), x)

    def ci(self, x, sigma=1):
        """
        Evaluate the polynomial and the confidence intervals at x.

        sigma=1 corresponds to a 1-sigma confidence interval
        """
        A = _poly_matrix(x, self.degree, self.origin)
        return self._conf.ci(A, sigma)

    def pi(self, x, p=0.05):
        """
        Evaluate the polynomial and the prediction intervals at x.

        p = 1-alpha = 0.05 corresponds to 95% prediction interval
        """
        A = _poly_matrix(x, self.degree, self.origin)
        return self._conf.pi(A, p)

    def __str__(self):
        # TODO: better polynomial pretty printing using formatnum
        return "Polynomial(%s)" % self.coeff


    def plot(self, ci=1, pi=0):
        import pylab
        min_x, max_x = np.min(self.x), np.max(self.x)
        padding = (max_x - min_x)*0.1
        x = np.linspace(min_x-padding, max_x+padding, 200)
        y = self.__call__(x)
        pylab.errorbar(self.x, self.y, self.dy, fmt='b.')
        pylab.plot(x, y, 'b-')
        if ci > 0:
            _, cdy = self.ci(x, ci)
            pylab.plot(x, y + cdy, 'b-.', x, y - cdy, 'b-.')
        if pi > 0:
            py, pdy = self.pi(x, pi)
            pylab.plot(x, y + pdy, 'b-.', x, y - pdy, 'b-.')


def wpolyfit(x, y, dy=1, degree=None, origin=False):
    r"""
    Return the polynomial of degree $n$ that minimizes $\sum(p(x_i) - y_i)^2/\sigma_i^2$.

    if origin is True, the fit should go through the origin.

    Returns :class:`PolynomialModel`.
    """
    assert degree is not None, "Missing degree argument to wpolyfit"

    A = _poly_matrix(x, degree, origin)
    s = wsolve(A, y, dy)
    return PolynomialModel(x, y, dy, s, origin=origin)


def smooth(x, xp, yp, dyp=None, degree=2, span=5):
    # type: (Sequence[float], Sequence[float], Sequence[float], Sequence[float], int, int) -> (np.ndarray, np.ndarray)
    """
    Moving least squares smoother.

    *x* are the points at which to evaluate the smoothed values, returning
    estimated *y +/- dy* for each point, where *dy* is the confidence interval
    for the value of *y*.

    *xp*, *yp*, *dyp* are the original data points and uncertainties.

    *degree* is the degree of the polynomial smoother, and *span* is the
    number of neighbouring points to use. If *xp* is equally spaced and
    evaluated at *x=xp*, then this is equivalent to a Savitsky-Golay filter
    of the same degree and span.  Linear interpolation is *degree=1* and
    *span=2*, but for that it will be much faster to use a linear interpolation
    function directly.

    Note: *xp* does not need to be equally spaced, but if it has large clumps
    then a fixed window length is not the most appropriate algorithm.
    Instead, one could specify a window width *dx* for each *x*, and use
    all data points within *[x-dx, x+dx]* as the input to the polynomial.
    This capability is not yet implemented.
    """
    if dyp is None:
        dyp = np.ones(len(yp))

    if len(xp) <= span:
        n = len(xp)
        if n <= degree:
            degree = n-1
        poly = wpolyfit(xp, yp, dyp, degree=degree)
        y, dy = poly.ci(x)

    else:
        y, dy = np.empty(x.shape), np.empty(x.shape)
        if span%2 == 0:
            # Even span is an odd number of intervals, so set boundaries
            # at the x points.
            index = np.searchsorted(xp[span//2:-span//2], x)
        else:
            # Odd span is even number of intervals, so set boundaries
            # at the midpoints between x.
            index = np.searchsorted(0.5*(xp[:-span]+xp[span:]), x)

        # Save a little when densely interpolating by caching poly
        # Since indices returned by searchsorted are always non-negative,
        # initializing last to -1 guarantees that the first time through the
        # loop last != start, and a new polynomial will be fit.
        last = -1
        for i, start in enumerate(index):
            if last != start:
                # Note that the centers are offset by -span//2 because the
                # search started that far into the xp array.
                s = slice(start, start+span)
                poly = wpolyfit(xp[s], yp[s], dyp[s], degree=degree)
            #if i%10 == 0:
            #    from ..wsolve import wpolyplot;wpolyplot(poly)
            y[i], dy[i] = [v[0] for v in poly.ci([y[i]])]

    return y, dy

def demo():
    """
    Fit a random cubic polynomial.
    """
    import pylab

    # Make fake data
    x = np.linspace(-15, 5, 15)
    th = np.polyval([.2, 3, 1, 5], x)  # polynomial
    dy = np.sqrt(np.abs(th))        # poisson uncertainty estimate
    y = np.random.normal(th, dy)    # but normal generator

    # Fit to a polynomial
    poly = wpolyfit(x, y, dy=dy, degree=3)
    poly.plot()
    pylab.show()

def demo2():
    import pylab
    x = [1, 2, 3, 4, 5]
    y = [10.2, 7.9, 6.9, 4.4, 1.8]
    dy = [1, 3, 1, 0.2, 1.5]
    poly = wpolyfit(x, y, dy=dy, degree=1)
    poly.plot()
    pylab.show()


def test():
    """
    Check that results are correct for a known problem.
    """
    from numpy.testing import assert_array_almost_equal_nulp
    x = np.array([0, 1, 2, 3, 4], 'd')
    y = np.array([2.5, 7.9, 13.9, 21.1, 44.4], 'd')
    dy = np.array([1.7, 2.4, 3.6, 4.8, 6.2], 'd')
    poly = wpolyfit(x, y, dy, 1)
    px = np.array([1.5], 'd')
    _, pi = poly.pi(px)  # Same y is returend from pi and ci
    py, ci = poly.ci(px)

    ## Uncomment these to show target values
    #print("    Tp = np.array([%.16g, %.16g])"%(tuple(poly.coeff.tolist())))
    #print("    Tdp = np.array([%.16g, %.16g])"%(tuple(poly.std.tolist())))
    #print("    Tpi, Tci = %.16g, %.16g"%(pi[0],ci[0]))
    Tp = np.array([7.787249069840739, 1.503992847461522])
    Tdp = np.array([1.522338103010216, 2.117633626902384])
    Tpi, Tci = 7.611128464981326, 2.34286039211232

    perr = np.max(np.abs(poly.coeff - Tp))
    dperr = np.max(np.abs(poly.std - Tdp))
    cierr = np.abs(ci - Tci)
    pierr = np.abs(pi - Tpi)
    assert perr < 1e-14, "||p-Tp||=%g" % perr
    assert dperr < 1e-14, "||dp-Tdp||=%g" % dperr
    # TODO: change target error level once scipy 1.13 becomes the norm
    # Target values are only accurate to 9 digits for scipy < 1.13
    assert cierr < 1e-8, "||ci-Tci||=%g" % cierr
    assert pierr < 1e-8, "||pi-Tpi||=%g" % pierr
    assert_array_almost_equal_nulp(py, poly(px), nulp=8)

if __name__ == "__main__":
    #test()
    demo()
    #demo2()



# This program is public domain
# Author: Paul Kienzle
"""
Dependency calculator.
"""
from __future__ import print_function

def processing_order(pairs, n=0):
    """
    Order the work in a workflow.

    Given a set of n items to evaluate numbered from zero through n-1,
    and dependency pairs


    :Parameters:

    *pairs* : [(int, int), ...]

        Pairwise dependencies amongst items.

    *n* : int

        Number of items, or 0 if we don't care about any item that is not
        mentioned in the list of pairs

    :Returns:

    *order* : [int, ...]

        Permutation which satisfies the partial order requirements.

    """
    order = _dependencies(pairs)
    if n:
        if any(id >= n for id in order):
            raise ValueError("Not all dependencies are in the set")
        rest = set(range(n)) - set(order)
    else:
        rest = set(k for p in pairs for k in p) - set(order)
    #print "order",order,"from",pairs
    return order + list(rest)


def _dependencies(pairs):
    # type: (List[Tuple[int, int]]) -> List[int]
    emptyset = set()
    order = []  # type: List[int]
    pairs = [(a, b) for a, b in pairs]
    if not pairs:
        return order

    # Break pairs into left set and right set
    left, right = (set(s) for s in zip(*pairs))
    while pairs:
        #print "within",pairs
        # Find which items only occur on the right
        independent = right - left
        if independent == emptyset:
            cycleset = ", ".join(str(s) for s in left)
            raise ValueError("Cyclic dependencies amongst %s" % cycleset)

        # The possibly resolvable items are those that depend on the independents
        dependent = set(a for a, b in pairs if b in independent)
        pairs = [(a, b) for a, b in pairs if b not in independent]
        if not pairs:
            resolved = dependent
        else:
            left, right = (set(s) for s in zip(*pairs))
            resolved = dependent - left
        #print "independent",independent,"dependent",dependent,"resolvable",resolved
        order += resolved
        #print("new order",order)
    order.reverse()
    return order


# ========= Test code ========
def _check(msg, pairs, n):
    """
    Verify that the list n contains the given items, and that the list
    satisfies the partial ordering given by the pairs in partial order.
    """
    order = processing_order(pairs, n=n)
    if len(set(order)) != n:
        raise RuntimeError("%s is missing items" % msg)
    for lo, hi in pairs:
        if order.index(lo) >= order.index(hi):
            raise RuntimeError("%s expect %s before %s in %s for %s"
                               % (msg, lo, hi, order, pairs))
    print("%s %s"%(msg, str(order)))


def test():
    import numpy as np

    # No dependencies
    _check("test empty", [], 9)

    # No chain dependencies
    _check("test2", [(4, 1), (3, 2), (7, 6)], 9)

    # Some chain dependencies
    pairs = [(4, 0), (0, 1), (1, 2), (7, 0), (3, 5)]
    _check("test1", pairs, 9)
    _check("test1 numpy", np.array(pairs), 9)

    # Cycle test
    pairs = [(1, 4), (4, 3), (4, 5), (5, 1)]
    try:
        _ = processing_order(pairs, n=9)
    except ValueError:
        pass
    else:
        raise Exception("test3 expect ValueError exception for %s" % (pairs,))

    # large test for gross speed check
    A = np.random.randint(4000, size=(1000, 2))
    A[:, 1] += 4000  # Avoid cycles
    _check("test-large", A, 8000)

    # depth tests
    k = 200
    A = np.array([range(0, k), range(1, k + 1)]).T
    _check("depth-1", A, 201)

    A = np.array([range(1, k + 1), range(0, k)]).T
    _check("depth-2", A, 201)


if __name__ == "__main__":
    test()



from .refldata import Intent, infer_intent

INTENT_OPTIONS = 'infer|auto|'+'|'.join(Intent.intents)
def apply_intent(data, intent):
    stored_intent = getattr(data, 'intent', Intent.none)
    inferred_intent = infer_intent(data)
    if stored_intent == Intent.none:
        data.intent = inferred_intent
    elif intent == 'infer':
        data.intent = inferred_intent
    elif intent == 'auto':
        pass  # data.intent already is stored_intent
    else:
        data.intent = intent
    if inferred_intent not in (data.intent, Intent.none, Intent.time):
        data.warn("intent %r does not match inferred intent %r"
                  % (data.intent, inferred_intent))





"""
Core class definitions
"""
from __future__ import print_function

import sys
import importlib
import inspect
import json
from collections import OrderedDict
from importlib import resources

from numpy import nan, inf

from .deps import processing_order

TEMPLATE_VERSION = '1.0'

_instrument_registry = OrderedDict()
_module_registry = {}
_datatype_registry = {}

_loaded_instruments = set() # previously loaded instruments
def load_instrument(name):
    """
    Load the dataflow instrument definition given by the instrument *name*.

    This is a convenience function that relies on a standard layout for
    the instrument definition.  It assumes there is a reduction package
    called *name+"red"* on the python path which contains the module
    *dataflow.py*.  The module should contain the symbol *INSTRUMENT*
    giving the instrument id (for example, "ncnr.refl") and the
    function *define_instrument()* to register the instrument if *INSTRUMENT*
    is not already registered.

    The *{name}red.dataflow.define_instrument()* method should define
    the :class:`Module` reduction steps, the :class:`DataType` items
    to pass information between them and an :class:`Instrument` object
    to pull them together.  It should call :func:`register_instrument`
    on the resulting instrument object.

    Unfortunately, the instrument definition module cannot be directly
    determined from a stored template. Each step in the template does
    define *"module": id* where *id* is a dotted name such as
    *"ncnr.refl.subtract_background"*. By convention the second element
    corresponds to the input *name* for this function.

    Note: You can call *load_instrument(name)* repeatedly with the same
    name, but you can't call it if you have called *define_instrument()*
    directly.  Attempting to do so will raise a ValueError stating that
    the instrument is already defined.

    TODO: Include instrument definition module name in the saved template.
    """
    module_name = f'reductus.{name}red.dataflow'
    if module_name not in _loaded_instruments:
        module = importlib.import_module(module_name)
        _loaded_instruments.add(module_name)
        # Note: need to check for INSTRUMENT even though we've already
        # checked for module_name because the module may have been loaded
        # the traditional way using something like:
        #     import myinstred; myinstred.define_instrument()
        if module.INSTRUMENT not in _instrument_registry:
            module.define_instrument()
        else:
            raise ValueError("Instrument {module.INSTRUMENT} already defined.".format(module=module))

def register_instrument(instrument):
    """
    Add a new instrument to the server.
    """
    _instrument_registry[instrument.id] = instrument
    for m in instrument.modules:
        register_module(m)
    for d in instrument.datatypes:
        register_datatype(d)


def lookup_instrument(id):
    """
    Find a predfined instrument given its id.
    """
    return _instrument_registry[id]


def list_instruments():
    """
    Return a list of available instruments.
    """
    return list(_instrument_registry.keys())


def register_module(module):
    """
    Register a new calculation module.
    """
    if module.id in _module_registry and module != _module_registry[module.id]:
        #raise TypeError("Module already registered")
        return
    _module_registry[module.id] = module


def lookup_module(id):
    """
    Lookup a module in the registry.
    """
    return _module_registry[id]


def register_datatype(datatype):
    if (datatype.id in _datatype_registry
            and datatype != _datatype_registry[datatype.id]):
        raise TypeError("Datatype already registered")
    _datatype_registry[datatype.id] = datatype


def lookup_datatype(id):
    return _datatype_registry[id]


def load_templates(package):
    """
    Returns a dictionary {name: template} for the given instrument.

    Templates are defined as JSON objects, with stored in a file named
    "<instrument>.<name>.json".  All templates for an instrument should
    be stored in a templates subdirectory, made into a package by inclusion
    of an empty __init__.py file.  They can then be loaded using::

        from reductus.dataflow import core as df
        from . import templates
        ...
        instrument = df.Instrument(
            ...
            templates=df.load_templates(templates),
        )
    """
    templates = {}
    templates_dir = resources.files(package)
    for item in templates_dir.iterdir():
        if item.is_file() and item.name.endswith('.json'):
            name = item.name.split('.')[-2]
            template = json.loads(templates_dir.joinpath(item.name).read_text())
            templates[name] = template
    return templates


class Module(object):
    """
    Processing module

    A computation is represented as a set of modules connected by wires.

    *id* : string
        Module identifier. By convention this will be a dotted structure
        '<instrument class>.<operation>.<qualifier>', with qualifier optional.
        For example, use "tas.load" for triple axis data and "sans.load" for
        SANS data.  For NCNR SANS data format loaders, use "sans.load.ncnr".

    *version* : string
        Version number of the code which implements the filter calculation.
        If any code in the supporting libraries changes in a way that will
        affect the calculation results, the version number should be
        incremented.  This includes bug fixes.

    *author* : string
        Author of the module

    *name* : string
        The display name of the module. This may appear in the user interface
        in addition to any pictorial representation of the module. Usually it
        is just the name of the operation. By convention, it should have
        every word capitalized, with spaces between words.

    *action* : callable
        function which performs the action

    *action_id* : string
        fully qualified name required to import the action

    *description* : string
        A tooltip shown when hovering over the icon

    *icon* : { URI: string, terminals: { *id*: [*x*,*y*,*i*,*j*] } }
        Image representing the module, or none if the module should be
        represented by name.

        The terminal locations are identified by:

            *id* : string
                name of the terminal

            *position* : [int, int]
                (x,y) location of terminal within icon

            *direction* : [int, int]
                direction of the wire as it first leaves the terminal;
                default is straight out

    *fields* : [{Parameter}, ...]
        A form defining the constants needed for the module. For
        example, an attenuator will have an attenuation scalar. Field
        names must be distinct from terminal names.

        *id* : string
            name of the variable associated with the data; this must
            correspond to a parameter name in the module action.

        *label* : string
            display name for the field.

        *type* : string
            name of the datatype associated with the data

        *default* : object
            default value if none specified in template

        *description* : string
            A tooltip shown when hovering over the field

        *required* : boolean
            true if a value is required

        *multiple* : boolean
            true if each input should get a different value when processing
            a bundle with the module.


    *inputs*, "outputs" : [{Terminal}, ...]
        List module inputs and outputs.

        *id* : string
            name of the variable associated with the data; this must
            correspond to a parameter name in the module action.

        *label* : string
            display name for the terminal.

        *type* : string
            name of the datatype associated with the data, with the
            output of one module needing to match the input of the
            next. Using a hierarchical type system, such as
            refl.data, we can attach to generic modules like scaling
            as well as specific modules like footprint correction. By
            defining the type of data that flows through the terminal
            we can highlight valid connections automatically.

        *description* : string
            A tooltip shown when hovering over the terminal; defaults
            to datatype name

        *required* : boolean
            true if an input is required; ignored on output terminals.

        *multiple* : boolean
            true if multiple inputs are accepted on input terminals,
            or if multiple outputs are produced on output terminals.
    """
    _source = None
    _terminal_by_id = None

    def __init__(self, id, version, name, description, icon=None,
                 inputs=None, outputs=None, fields=None, action=None,
                 author="", action_id="",
                ):
        self.id = id
        self.version = version
        self.name = name
        self.description = description
        self.author = author
        self.icon = icon
        self.fields = fields if fields is not None else {}
        self.inputs = inputs
        self.outputs = outputs
        self.action = action
        self.action_id = action_id

    def get_terminal_by_id(self, id):
        """
        Lookup terminal by id, and return.
        Returns None if id does not exist.
        """
        if self._terminal_by_id is None:
            self._terminal_by_id = dict((t['id'], t)
                                        for t in self.inputs+self.outputs)
        return self._terminal_by_id[id]

    def get_source_code(self):
        """
        Retrieves the source code for the identified module that
        does the actual calculation.  If no module is identified
        it returns an empty string
        """
        if self._source is None:
            self._source = "".join(inspect.getsourcelines(self.action)[0])
        return self._source

    def get_definition(self):
        return self.__getstate__()

    @property
    def cached(self):
        return not hasattr(self.action, 'cached') or self.action.cached

    @property
    def visible(self):
        return not hasattr(self.action, 'visible') or self.action.visible

    def __getstate__(self):
        # Don't pickle the function reference
        keys = ['version', 'id', 'name', 'description', 'icon',
                'fields', 'inputs', 'outputs', 'action_id', 'visible']
        return dict([(k, getattr(self, k)) for k in keys])

    def __setstate__(self, state):
        from importlib import import_module
        for k, v in state.items():
            setattr(self, k, v)
        # Restore the function reference after unpickling
        parts = self.action_id.split('.')
        mod = import_module(".".join(parts[:-1]))
        self.action = getattr(mod, parts[-1])


class Template(object):
    """
    A template captures the computational workflow as a wiring diagram.

    *name* : string
        String identifier for the template

    *description* : string
        Extended description to be displayed as help to the template user.

    *modules* : [{TemplateModule}, ...]
        Modules used in the template

        *module* : string
            module id for template node

        *version* : string
            version number of the module

        *config* : { field: value }
            initial values for the fields

        *position* : [int,int]
            location of the module on the canvas.

    *wires* : [{TemplateWire}, ...]
        Wires connecting the modules

        *source* : [int, string]
            module id in template and terminal name in module

        *target* : [int, string]
            module id in template and terminal name in module

    *instrument* : string
        Instrument to which the template applies

    *version* : string
        Template version number
    """
    def __init__(self, name=None, description=None, modules=None, wires=None,
                 instrument="", version=TEMPLATE_VERSION, **kw):
        self.name = name
        self.description = description
        self.modules = modules
        self.wires = wires
        self.instrument = instrument
        self.version = version

    def order(self, target=None):
        """
        Return the module ids in processing order.
        """
        if target is None:
            # Order to evaluate all nodes
            pairs = [(w['source'][0], w['target'][0]) for w in self.wires]
            n = len(self.modules)
        else:
            pairs = []
            n = 0
            remaining = set([target])
            processed = set()
            while remaining:
                target = remaining.pop()
                sources = [w['source'][0] for w in self.inputs(target)]
                pairs.extend((source, target) for source in sources)
                processed.add(target)
                remaining |= set(sources) - processed
            if not pairs:  # No dependencies; calculate node only
                return [target]
        return processing_order(pairs, n)

    def dependents(self, id):
        """
        Retrieve the list of nodes that depend on a particular node, including
        the node itself.

        Note: Algorithm is O(n^2) or worse, so be prepared to rewrite if
        performance is an issue.
        """
        #print "wires", [(w['source'][0], w['target'][0]) for w in self.wires]
        remaining = set([id])
        processed = set([id])
        while remaining:
            # pick an unprocessed node
            parent = remaining.pop()
            # find which nodes depend on it
            children = set(w['target'][0] for w in self.wires
                           if w['source'][0] == parent)
            # remember to process those that are not already listed
            remaining |= children - processed
            # list the new nodes as descendents
            processed |= children
            #print "deps",id,parent,children,remaining,processed
        return processed

    def inputs(self, id):
        """
        Retrieve the data objects that go into the inputs of a module
        """
        wires = [w for w in self.wires if w['target'][0] == id]
        return wires

    def ordered(self, target=None):
        """
        Yields (module, inputs) for each module in evaluation order.

        If *target* is specified, only include modules required to
        evaluate the target.
        """
        for id in self.order(target):
            yield id, self.inputs(id)

    def dumps(self, **kw):
        """
        Convert template to json.
        """
        return json.dumps(self.__getstate__(), **kw)

    def show(self):
        """
        Print template on console as prettified json.
        """
        print(json.dumps(self.__getstate__(), indent=2, sort_keys=True,
                         separators=(',', ': ')))

    def get_definition(self):
        return self.__getstate__()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        # As the template definition changes we need to increment the version
        # number in TEMPLATE_VERSION.  This code must be able to interpret
        # older versions of the template; it should never need to interpret
        # newer versions of the template, but should still protect against it.
        # In the initial release, clearly the the template version must
        # match TEMPLATE_VERSION since there is only one version.
        if state['version'] != TEMPLATE_VERSION:
            raise TypeError('Template definition mismatch')
        self.__dict__ = state


class Instrument(object):
    """
    An instrument is a set of modules and standard templates to be used
    for reduction

    *id* : string
        Instrument identifier. By convention this will be a dotted
        structure '<facility>.<instrument class>.<instrument>'

    *name* : string
        The display name of the instrument

    *menu* : [(string, [Module, ...]), ...]
        Modules available. Modules are organized into groups of related
        operations, such as Input, Reduce, Analyze, ...

    *datatypes* : [Datatype]
        List of datatypes used by the instrument

    *archive* : URI
        Location of the data archive for the instrument. Archives must
        implement an interface that allows data sets to be listed and
        retrieved for a particular instrument/experiment.
    """
    def __init__(self, id, name=None, menu=None, template_defs=None,
                 datatypes=None, requires=None, archive=None, loaders=None):
        self.id = id
        self.name = name
        self.menu = menu
        self.template_defs = template_defs if template_defs is not None else []
        self.datatypes = datatypes
        self.requires = requires
        self.archive = archive

        self.loaders = loaders

        self.modules = []
        for _, m in menu:
            self.modules.extend(m)
        self._check_datatypes()
        self._check_names()

    def _check_datatypes(self):
        defined = set(d.id for d in self.datatypes)
        used = set()
        for m in self.modules:
            used |= set(t['datatype'] for t in m.inputs+m.outputs)
        if used - defined:
            raise TypeError("undefined types: %s" % ", ".join(used - defined))
        if defined - used:
            #raise TypeError("unused types: %s" % ", ".join(defined - used))
            print("Warning: unused types: %s" % ", ".join(defined - used))

    def _check_names(self):
        names = set(m.name for m in self.modules)
        if len(names) != len(self.modules):
            raise TypeError("names must be unique within an instrument")

    def get_module_by_id(self, id):
        if '.' not in id:
            id = ".".join((self.id, id))
        for m in self.modules:
            #print "checking %r against %r"%(m.id, id)
            if m.id == id:
                return m
        else:
            raise KeyError(id + ' does not exist in instrument ' + self.name)

    def get_module_by_name(self, name):
        for m in self.modules:
            if m.name == name:
                return m
        else:
            raise KeyError(name + ' does not exist in instrument ' + self.name)

    def id_by_name(self, name):
        for m in self.modules:
            if m.name == name: return m.id
        raise KeyError(name + ' does not exist in instrument ' + self.name)

    def get_definition(self):
        keys = ['id', 'name', 'archive']
        definition = dict([(k, getattr(self, k)) for k in keys])
        definition['modules'] = [m.get_definition() for m in self.modules]
        definition['datatypes'] = [d.get_definition() for d in self.datatypes]
        definition['templates'] = self.template_defs
        return definition


class DataType(object):
    """
    Data objects represent the information flowing over a wire.

    *id* : string
        Name of the data type.

    *cls* : Class
        Python class defining the data type.

    **Data Visualization**

    Data objects are displayed to the user, with the display format
    returned by *data.get_plottable()*.  The details of the plottable
    object are still fluid, and defined by the webreduce server.  See
    web_gui/static/js/webreduce/editor.js for the implementation.

    **Metadata**

    Information about the data object is sent to the web browser client via
    the *data.get_metadata()* method.

    **TODO**: add details about the structure and use of metadata.  It looks
    to be a simple JSON mapping of the data object which is available on
    the debug console of the browser.  Check if the webreduce client uses
    any specific fields from the object.

    **Data Export**

    Data objects define available exporters as::

        from reductus.dataflow.lib import exporters
        self.export_types = {
            "NAME" : {
                "method_name": "METHOD",
                "exporter": exporters.COMPOSITOR,
            }
        }

    where NAME is export type shown to the user, METHOD is the name of the
    data method which builds the exported object, and COMPOSITOR is
    a function that knows how to bundle entries into independent or
    concatenated export files, each with the template metadata attached.

    The export types are discovered by introspection, using a decorator
    on the entry constructor such as *@exporters.exports_HDF5("NAME")*.

    The constructor method needs to return an object of the form::

        {
            "name": "NAME",
            "entry": "ENTRY",
            "file_suffix": ".EXT",
            "value": formatted_data,
        }

    The resulting data will be saved in *NAME_ENTRY.EXT*.  For concatenated
    results the filename for the first item will be used.
    """
    def __init__(self, id, cls):
        self.id = id
        self.cls = cls
        self.export_types = self._find_exporters()

    def get_definition(self):
        return {"id": self.id, "export_types": list(self.export_types.keys())}

    def _find_exporters(self):
        exporters = {}
        for method_name in dir(self.cls):
            method = getattr(self.cls, method_name)
            if not callable(method):
                continue
            exporter = getattr(method, 'exporter', None)
            if exporter is not None:
                export_name = getattr(method, 'export_name', 'default')
                exporters[export_name] = {
                    "exporter": exporter,
                    "method_name": method_name,
                }
        return exporters

class Bundle(object):
    def __init__(self, datatype, values):
        self.datatype = datatype
        self.values = values

    def todict(self):
        values = [v.todict() for v in self.values]
        return {'datatype': self.datatype.id, 'values': values}

    def get_plottable(self):
        values = [v.get_plottable() for v in self.values]
        return {'datatype': self.datatype.id, 'values': values}

    def get_metadata(self):
        values = [v.get_metadata() for v in self.values]
        return {'datatype': self.datatype.id, 'values': values}

    def get_export(self, export_type="column", template_data=None, concatenate=True):
        if export_type not in self.datatype.export_types:
            raise ValueError("{self.datatype.id} does not provide {export_type} export.".format(self=self, export_type=export_type))
        exporter_info = self.datatype.export_types[export_type]
        exporter = exporter_info["exporter"]
        to_export = exporter(
            self.values,
            export_method=exporter_info["method_name"],
            template_data=template_data,
            concatenate=concatenate)
        return {'datatype': self.datatype.id, 'values': to_export}

    @staticmethod
    def fromdict(state):
        datatype = lookup_datatype(state['datatype'])
        values = []
        for value in state['values']:
            obj = datatype.cls()
            obj.fromdict(value)
            values.append(obj)
        return Bundle(datatype, values)


# Inf/NaN representation options:
#     javascript names: "Infinity", "-Infinity", "NaN"
#     python names: "inf", "-inf", "nan"
#     math expressions: "1/0", "-1/0", "0/0"
#     unicode symbols: u"\u221E", u"-\u221E", u"\u26A0"
#     u26A0 is WARNING SIGN (! in triangle)
#     uFFFD is REPLACEMENT CHARACTER (? in diamond)
_NAN_STRING = u"\u26A0"  # WARNING SIGN (! in triangle)
_INF_STRING = u"\u221E"  # INFINITY
_MINUS_INF_STRING = u"-\u221E"  # -INFINITY
def sanitizeForJSON(obj):
    """
    Take an object made of python objects and remove inf and nan
    """
    if isinstance(obj, dict):
        output = {}
        for k, v in obj.items():
            output[k] = sanitizeForJSON(v)
        return output
    elif isinstance(obj, list):
        return list(map(sanitizeForJSON, obj))
    elif obj == inf:
        # Use WARNING SIGN for NaN
        return _INF_STRING
    elif obj == -inf:
        return _MINUS_INF_STRING
    elif obj != obj:
        # Use WARNING SIGN for NaN
        return _NAN_STRING
    else:
        return obj


def sanitizeFromJSON(obj):
    """
    Convert inf/nan from json representation to python.
    """
    if isinstance(obj, dict):
        output = {}
        for k, v in obj.items():
            output[k] = sanitizeFromJSON(v)
        return output
    elif isinstance(obj, list):
        return map(sanitizeFromJSON, obj)
    elif obj == _INF_STRING:
        return inf
    elif obj == _MINUS_INF_STRING:
        return -inf
    elif obj == _NAN_STRING:
        return nan
    else:
        return obj




from __future__ import print_function

import os
from pprint import pprint
import json
import traceback

from reductus import dataflow
from reductus.dataflow.core import Template, load_instrument, lookup_instrument
from reductus.dataflow.core import list_instruments as _list_instruments
from reductus.dataflow.cache import get_cache
from reductus.dataflow.calc import process_template
from reductus.rev import revision_info
from reductus.dataflow import configure
from reductus.dataflow import fetch

api_methods = []

def expose(action):
    """
    Decorator which adds function to the list of methods to expose in the api.
    """
    api_methods.append(action.__name__)
    return action

def sorted_ls(path, show_hidden=False):
    mtime = lambda f: os.stat(os.path.join(path, f)).st_mtime
    return list(sorted(filter(lambda x: os.path.exists(os.path.join(path, x)) and not x.startswith("."), os.listdir(path)), key=mtime))

def local_file_metadata(pathlist):
    # only absolute paths are supported:
    path = os.path.join(os.sep, *pathlist)
    dirlisting = sorted_ls(path)
    subdirs = []
    files = []
    files_metadata = {}
    for di in dirlisting:
        d = os.path.join(path, di)
        if os.path.isdir(d):
            subdirs.append(di)
        elif os.path.isfile(d):
            files.append(di)
            files_metadata[di] = {"mtime": int(os.path.getmtime(d))}
        else:
            # you've probably hit an unfulfilled path link or something.
            pass

    metadata = {
        "subdirs": subdirs,
        "files": files,
        "pathlist": pathlist,
        "files_metadata": files_metadata
        }
    return metadata

@expose
def get_file_metadata(source="ncnr", pathlist=None):
    if pathlist is None:
        pathlist = []

    if source not in [s['name'] for s in fetch.DATA_SOURCES]:
        raise ValueError("Source '{source}' not in available data sources".format(source=source))
    if source == "local":
        metadata = local_file_metadata(pathlist)
    else:
        import requests
        url = fetch.FILE_HELPERS[source] #'https://ncnr.nist.gov/ipeek/listftpfiles_json.php'
        req = requests.post(url, json={"pathlist": pathlist})
        metadata = req.json()

    return metadata

@expose
def get_instrument(instrument_id="ncnr.refl"):
    """
    Make the instrument definition available to clients
    """
    instrument = lookup_instrument(instrument_id)
    return instrument.get_definition()

def refl_load(file_descriptors):
    """
    file_descriptors will be a list of dicts like
    [{"path": "ncnrdata/cgd/201511/21066/data/HMDSO_17nm_dry14.nxz.cgd", "mtime": 1447353278}, ...]
    """
    modules = [{"module": "ncnr.refl.load", "version": "0.1", "config": {}}]
    template = Template("test", "test template", modules, [], "ncnr.magik", version='0.0')
    retval = process_template(template, {0: {"files": file_descriptors}}, target=(0, "output"))
    return retval.todict()

@expose
def find_calculated(template_def, config):
    """
    Returns a vector of true/false for each node in the template indicating
    whether that node value has been calculated yet.
    """
    template = Template(**template_def)
    retval = dataflow.calc.find_calculated(template, config)
    return retval

@expose
def calc_terminal(template_def, config, nodenum, terminal_id, return_type='full', export_type="column", concatenate=True):
    """ json-rpc wrapper for calc_single
    template_def =
    {"name": "template_name",
     "description": "template description!",
     "modules": ["list of modules"],
     "wires": ["list of wires"],
     "instrument": "facility.instrument_name",
     "version": "2.7.3"
    }

    where modules in list of modules above have structure:
    module =
    {"module": "facility.instrument_name.module_name",
     "version": "0.3.2"
    }

    and wires have structure:
    [["wire_start_module_id:wire_start_terminal_id", "wire_end_module_id:wire_end_terminal_id"],
     ["1:output", "2:input"],
     ["0:xslice", "3:input"]
    ]

    config =
    [{"param": "value"}, ...]

    nodenum is the module number from the template for which you wish to get the calculated value

    terminal_id is the id of the terminal for that module, that you want to get the value from
    (output terminals only).
    """
    template = Template(**template_def)
    #print "template_def:", template_def, "config:", config, "target:",nodenum,terminal_id
    #print "modules","\n".join(m for m in df._module_registry.keys())
    try:
        retval = process_template(template, config, target=(nodenum, terminal_id))
    except Exception:
        print("==== template ===="); pprint(template_def)
        print("==== config ===="); pprint(config)
        traceback.print_exc()
        raise
    if return_type == 'full':
        return retval.todict()
    elif return_type == 'plottable':
        return retval.get_plottable()
    elif return_type == 'metadata':
        return retval.get_metadata()
    elif return_type == 'export':
        # inject git version hash into export data:
        rev_id = revision_info()
        template_data = {
            "template_data": {
                "template": template_def,
                "config": config,
                "node": nodenum,
                "terminal": terminal_id,
                "server_git_hash": rev_id,
                "export_type": export_type,
                #"datasources": fetch.DATA_SOURCES, # Is this needed?
            }
        }
        to_export = retval.get_export(
            export_type=export_type, template_data=template_data,
            concatenate=concatenate)

        return to_export

    raise KeyError(return_type + " not a valid return_type (should be one of ['full', 'plottable', 'metadata', 'export'])")

@expose
def calc_template(template_def, config):
    """ json-rpc wrapper for process_template """
    template = Template(**template_def)
    #print "template_def:", template_def, "config:", config
    try:
        retvals = process_template(template, config, target=(None, None))
    except Exception:
        print("==== template ===="); pprint(template_def)
        print("==== config ===="); pprint(config)
        #traceback.print_exc()
        raise
    output = {}
    for rkey, rv in retvals.items():
        module_id, terminal_id = rkey
        module_key = str(module_id)
        output.setdefault(module_key, {})
        output[module_key][terminal_id] = rv.todict()
    return output

@expose
def list_datasources():
    return fetch.DATA_SOURCES

@expose
def list_instruments():
    return _list_instruments()

def initialize(config=None):
    if config is None:
        config = configure.load_config('config')
    configure.apply_config(user_config=config)

if __name__ == '__main__':
    initialize()



from reductus.dataflow import core as df
from reductus.dataflow.automod import make_modules, make_template, auto_module, get_modules
from reductus.dataflow.calc import process_template
from reductus.dataflow.data import Plottable
from reductus.dataflow.lib.exporters import exports_json

from . import steps
from . import candor_steps
from . import templates
from .refldata import ReflData, PSDData
from .candor import Candor
from .polarization import PolarizationData
from .deadtime import DeadTimeData
from .footprint import FootprintData
from .backgroundfield import BackgroundFieldData

INSTRUMENT = "ncnr.refl"

class FluxData(object):
    def __init__(self, fluxes, total_flux):
        self.fluxes = fluxes
        self.total_flux = total_flux

    def get_metadata(self):
        return {
            "fluxes": self.fluxes,
            "total_flux": self.total_flux
        }

    def get_plottable(self):
        return self.get_metadata()

    @exports_json
    def to_json_text(self):
        return {
            "name": "fluxes",
            "entry": "",
            "file_suffix": ".json",
            "value": self.get_metadata(),
            "compact": False,  # pretty-print result
        }

def make_cached_subloader_module(load_action, prefix=""):
    """
    This assumes that the load_action can be run with a single fileinfo
    in any of the fields of datatype 'fileinfo',
    and collates the results of running them one at a time.
    """
    # Read the module defintion from the docstring
    module_description = auto_module(load_action)
    fields = module_description['fields']
    fileinfo_fields = [f for f in fields if f['datatype'] == 'fileinfo']

    # Tag module ids with prefix
    module_description['name'] += " (cached)"
    mod_id = module_description['id'] = prefix + module_description['id']
    template_def = {
        "name": "loader_template",
        "description": "cached remote loader",
        "modules": [
            {"module": mod_id, "version": "0.1", "config": {}}
        ],
        "wires": [],
        "instrument": INSTRUMENT,
        "version": "0.0"
    }
    template = df.Template(**template_def)

    # Tag each terminal data type with the data type prefix, if it is
    # not already a fully qualified name
    for v in module_description['inputs'] + module_description['outputs']:
        if '.' not in v['datatype']:
            v['datatype'] = prefix + v['datatype']

    def new_action(**kwargs):
        outputs = []
        fileinfos = {}
        for ff in fileinfo_fields:
            fileinfos[ff['id']] = kwargs.pop(ff['id'], [])
            # replace fileinfos with empty lists
            kwargs[ff['id']] = []
        for field_id in fileinfos:
            fileinfo = fileinfos[field_id]
            for fi in fileinfo:
                # put a single fileinfo into the hopper
                kwargs[field_id] = [fi]
                config = {"0": kwargs}
                nodenum = 0
                terminal_id = "output"
                retval = process_template(template, config, target=(nodenum, terminal_id))
                outputs.extend(retval.values)
                # take it back out before continuing the loop(s)
                kwargs[field_id] = []
        return outputs

    new_action.cached = True
    module_description['id'] += ".cached"
    # Define and register the module
    return df.Module(action=new_action, **module_description)

def define_instrument():
    # Define modules
    actions = get_modules(steps) + get_modules(candor_steps)
    modules = make_modules(actions, prefix=INSTRUMENT+'.')
    modules.append(make_cached_subloader_module(steps.super_load, prefix=INSTRUMENT+'.'))
    modules.append(make_cached_subloader_module(steps.ncnr_load, prefix=INSTRUMENT+'.'))
    # Define data types
    refldata = df.DataType(INSTRUMENT+".refldata", ReflData)
    poldata = df.DataType(INSTRUMENT+".poldata", PolarizationData)
    psddata = df.DataType(INSTRUMENT+".psddata", PSDData)
    candordata = df.DataType(INSTRUMENT+".candordata", Candor)
    deadtime = df.DataType(INSTRUMENT+".deadtime", DeadTimeData)
    footprint = df.DataType(INSTRUMENT+".footprint.params", FootprintData)
    flux = df.DataType(INSTRUMENT+".flux.params", FluxData)
    backgroundfield = df.DataType(INSTRUMENT + ".backgroundfield.params", BackgroundFieldData)
    plottable = df.DataType(INSTRUMENT+".plot", Plottable)

    #import json
    #import os
    #from pkg_resources import resource_string, resource_listdir
    #template_names = [fn for fn in resource_listdir('dataflow', 'templates') if fn.endswith(".json")]
    #templates = [json.loads(resource_string('dataflow', 'templates/' + tn)) for tn in template_names]
    #template_path = resource_path("../dataflow/templates")
    #template_names = [fn for fn in os.listdir(template_path) if fn.endswith(".json")]
    #templates = [json.loads(open(os.path.join(template_path, tn), 'r').read()) for tn in template_names]

    # Define instrument
    refl1d = df.Instrument(
        id=INSTRUMENT,
        name='NCNR reflectometer',
        menu=[('steps', modules)],
        datatypes=[
            refldata, poldata, psddata, candordata, deadtime, footprint,
            flux, backgroundfield, plottable,
            ],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(refl1d)
    return refl1d


def loader_template():
    refl1d = df.lookup_instrument(INSTRUMENT)
    diagram = [
        ["ncnr_load", {}],
        ["divergence", {"data": "-.output"}],
    ]
    template = make_template(
        name="loader",
        description="loader only",
        diagram=diagram,
        instrument=refl1d,
        version=1.0,
        )
    return template


def unpolarized_template():
    refl1d = df.lookup_instrument(INSTRUMENT)
    diagram = [
        # Load the data
        ["ncnr_load", {}],

        # Preprocessing common to all data sets
        #["monitor_saturation", {"data": "-.output"}],
        #["detector_saturation", {"data": "-.output"}],
        ["divergence", {"data": "-.output"}],
        ["normalize", {"data": "-.output", "base": "auto"}],
        ["mark_intent", {"data": "-.output", "intent": ["auto"]}],
        ["group_by_intent => split", {"data": "-.output"}],

        # Preprocessing particular to each data type
        # Use nop to restart pipeline after split so that it is easier to
        # add/remove/rearrange steps for the different data types.
        ["nop", {"data": "split.specular"}],
        ["join => spec", {"data": "-.output"}],

        ["nop", {"data": "split.backp"}],
        ["mask_specular", {"data": "-.output"}],
        ["align_background", {"data": "-.output", "offset": "auto"}],
        ["join => backp", {"data": "-.output"}],

        ["nop", {"data": "split.backm"}],
        ["mask_specular", {"data": "-.output"}],
        ["align_background", {"data": "-.output", "offset": "auto"}],
        ["join => backm", {"data": "-.output"}],

        ["nop", {"data": "split.intensity"}],
        ["rescale", {"data": "-.output", "scale": [1.0], "dscale": [0.0]}],
        ["join => intensity", {"data": "-.output", "tolerance": 0.0001}],

        # Operate on the combined data for the final reduction
        ["subtract_background", {
            "data": "spec.output",
            "backp": "backp.output",
            "backm": "backm.output"}],
        ["divide_intensity", {"data": "-.output", "base": "intensity.output"}],
        #["footprint",  {"data": "-.output")}],
    ]
    #for m in refl1d.modules: print m.__dict__
    #for m in refl1d.modules: print m.terminals[0]
    template = make_template(
        name="unpolarized",
        description="standard unpolarized reduction",
        diagram=diagram,
        instrument=refl1d,
        version=1.0,
        )
    return template

def demo():
    from reductus.dataflow.calc import process_template

    template = unpolarized_template()
    #print "========== Template ========"
    #template.show()
    #print "="*24
    test_dataset = [{'path': "ncnrdata/cgd/201511/21066/data/HMDSO_17nm_dry14.nxz.cgd",
                     'mtime': 1447353278}]
    refl = process_template(template=template,
                            config={"0": {"filelist": test_dataset}},
                            target=(len(template.modules)-1, "output"),
                            #target=(2, "data"),  # return an input
                           )
    #print "refl",refl.values
    return refl.values


if __name__ == "__main__":
    from reductus.dataflow.cache import use_redis
    use_redis()
    define_instrument()
    demo()




import io
from zipfile import ZipFile, is_zipfile

import h5py

from . import hzf_readonly_stripped as hzf

def h5_open_zip(filename, file_obj=None, **kw):
    """
    Open a NeXus file, even if it is in a zip file,
    or if it is a NeXus-zip file.

    If the filename ends in '.zip', it will be unzipped to a temporary
    directory before opening and deleted on :func:`closezip`.  If opened
    for writing, then the file will be created in a temporary directory,
    then zipped and deleted on :func:`closezip`.

    If it is a zipfile but doesn't end in '.zip', it is assumed
    to be a NeXus-zip file and is opened with that library.

    Arguments are the same as for :func:`open`.
    """
    if file_obj is None:
        file_obj = io.BytesIO(open(filename, mode='rb', buffering=-1).read())
    is_zip = is_zipfile(file_obj) # is_zipfile(file_obj) doens't work in py2.6
    if is_zip and '.attrs' in ZipFile(file_obj).namelist():
        # then it's a nexus-zip file, rather than
        # a zipped hdf5 nexus file
        f = hzf.File(filename, file_obj)
    else:
        if is_zip:
            zf = ZipFile(file_obj)
            members = zf.namelist()
            assert len(members) == 1
            file_obj = io.BytesIO(zf.read(members[0]))
            filename = members[0]

        f = h5py.File(file_obj, **kw)
    return f



"""
SANS data loader
================

Load SANS NeXus file into :mod:`sansred.sansdata` data structure.
"""
import io
from zipfile import ZipFile, is_zipfile
from collections import OrderedDict
import h5py

from reductus.dataflow.lib import hzf_readonly_stripped as hzf
from reductus.dataflow.lib import unit
from reductus.dataflow.lib.h5_open import h5_open_zip

from reductus.vsansred.loader import load_detector, load_metadata
from reductus.vsansred.steps import _s, _b

from .sansdata import SansData, RawSANSData

metadata_lookup = OrderedDict([
    ("run.filename", "DAS_logs/trajectoryData/fileName"),
    ("analysis.intent", "DAS_logs/trajectoryData/intent"),
    ("analysis.filepurpose", "DAS_logs/trajectoryData/filePurpose"),
    ("det.des_dis", "DAS_logs/detectorPosition/desiredSoftPosition"),
    ("det.dis", "DAS_logs/detectorPosition/softPosition"),
    ("run.guide", "DAS_logs/guide/guide"),
    ("sample.description", "DAS_logs/sample/description"), # overwritten on load
    ("sample.labl", "DAS_logs/sample/description"), # compatibility
    ("resolution.lmda", "instrument/monochromator/wavelength"),
    ("resolution.dlmda", "instrument/monochromator/wavelength_error"),
    ("det.beamx", "instrument/detector/beam_center_x"),
    ("det.beamy", "instrument/detector/beam_center_y"),
    ("det.bstop", "DAS_logs/beamStop/size"),
    #("det.pixeloffsetx", "instrument/detector/x_offset"), # this is useless
    ("det.pixelsizex", "instrument/detector/x_pixel_size"),
    #("det.pixeloffsety", "instrument/detector/y_offset"), # this is also useless
    ("det.pixelsizey", "instrument/detector/y_pixel_size"),
    ("sample.name", "DAS_logs/sample/name"),
    ("polarization.front", "DAS_logs/frontPolarization/direction"),
    ("polarization.back", "DAS_logs/backPolarization/direction"),
    ("polarization.backname", "DAS_logs/backPolarization/name"),
    ("run.filePrefix", "DAS_logs/trajectoryData/filePrefix"),
    ("run.instFileNum", "DAS_logs/trajectoryData/instFileNum"),
    ("run.experimentScanID", "DAS_logs/trajectory/experimentScanID"),
    ("run.instrumentScanID", "DAS_logs/trajectory/instrumentScanID"),
    ("run.experimentPointID", "DAS_logs/trajectory/experimentPointID"),
    ("run.pointnum", "DAS_logs/trajectoryData/pointNum"),
    ("run.detcnt", "control/detector_counts"),
    ("run.rtime", "control/count_time"),
    ("run.moncnt", "control/monitor_counts"),
    ("run.atten", "DAS_logs/counter/actualAttenuatorsDropped"),
    ("analysis.groupid", "DAS_logs/trajectoryData/groupid"),
    ("run.configuration", "DAS_logs/configuration/key"),
    ("sample.thk", "DAS_logs/sample/thickness"),
    ("adam.voltage", "DAS_logs/adam4021/voltage"),
    ("sample.temp", "DAS_logs/temp/primaryNode/average_value"),
    ("resolution.ap1", "DAS_logs/geometry/sourceAperture"),
    ("resolution.ap2", "DAS_logs/geometry/externalSampleAperture"),
    ("resolution.ap2Off", "instrument/sample_aperture/distance"),
    ("resolution.ap12dis", "DAS_logs/geometry/sourceApertureToSampleAperture"),
    ("sample.position", "DAS_logs/geometry/samplePositionOffset"),
    ("electromagnet_lrm.field","DAS_logs/electromagnet_lrm/field"),
    ("mag.value","DAS_logs/mag/value"),
    ("acamplitude.voltage",  "DAS_logs/acAmplitude/voltage"),
    ("waveformgenerator.frequency",  "DAS_logs/waveformGenerator/frequency"),
    ("rfflipperpowersupply.voltage",  "DAS_logs/RFFlipperPowerSupply/actualVoltage/average_value"),
    ("rfflipperpowersupply.frequency",  "DAS_logs/RFFlipperPowerSupply/frequency"),
    ("huberRotation.softPosition",  "DAS_logs/huberRotation/softPosition"),
    ("start_time","start_time"),
    ("end_time","end_time"),
    ("eventfile", "DAS_logs/areaDetector/eventFileName"),
])

unit_specifiers = {
    "det.dis": "cm",
    "det.pixelsizex": "cm",
    "det.pixeloffsetx": "cm",
    "det.pixelsizey": "cm",
    "det.pixeloffsety": "cm",
    "sample.thk": "cm",
    "resolution.ap1": "cm",
    "resolution.ap2": "cm",
    "resolution.ap2Off": "cm",
}

def process_sourceAperture(field, units):
    import numpy as np
    def handler(v):
        return float(v.split()[0])
    handle_values = np.vectorize(handler)
    value = handle_values(field[()])
    units_from = ""
    v0 = field[0].split()
    if len(v0) > 1:
        units_from = v0[1]
    if type(units_from) == bytes:
        units_from = units_from.decode('utf-8')
    converter = unit.Converter(units_from)
    return converter(value, units)    

def data_as(field, units):
    """
    Return value of field in the desired units.
    """
    if field.name.split('/')[-1] == 'sourceAperture':
        return process_sourceAperture(field, units)
    else:
        units_in = field.attrs.get('units', '')
        if type(units_in) == bytes:
            units_in = units_in.decode('utf-8')
        converter = unit.Converter(units_in)
        value = converter(field[()], units)
        return value

def readSANSNexuz(input_file, file_obj=None, metadata_lookup=metadata_lookup):
    """
    Load all entries from the NeXus file into sans data sets.
    """
    datasets = []
    file = h5_open_zip(input_file, file_obj)
    for entryname, entry in file.items():        
        multiplicity = 1
        for i in range(multiplicity):
            metadata = load_metadata(entry, multiplicity, i, metadata_lookup=metadata_lookup, unit_specifiers=unit_specifiers)
            #print(metadata)
            detector_keys = ['detector']
            detectors = dict([(k, load_detector(entry['instrument'][k])) for k in detector_keys])
            metadata['entry'] = entryname
            # hack to remove configuration from sample label (it is still stored in run.configuration)
            metadata['sample.description'] = _s(metadata["sample.labl"]).replace(_s(metadata["run.configuration"]), "")
            dataset = RawSANSData(metadata=metadata, detectors=detectors)
            datasets.append(dataset)            

    return datasets

def readSANSNexuz_old(input_file, file_obj=None):
    """
    Load all entries from the NeXus file into sans data sets.
    """
    datasets = []
    file = h5_open_zip(input_file, file_obj)
    for entryname, entry in file.items():
        areaDetector = entry['data/areaDetector'][()]
        shape = areaDetector.shape
        if len(shape) < 2 or len(shape) > 3:
            raise ValueError("areaDetector data must have dimension 2 or 3")
            return
        if len(shape) == 2:
            # add another dimension at the front
            shape = (1,) + shape
            areaDetector = areaDetector.reshape(shape)
            
        for i in range(shape[0]):
            metadata = {}
            for mkey in metadata_lookup:
                field = entry.get(metadata_lookup[mkey], None)
                if field is not None:
                    if mkey in unit_specifiers:
                        field = data_as(field, unit_specifiers[mkey])
                    else:
                        field = field[()]
                    if field.dtype.kind == 'f':
                        field = field.astype("float")
                    elif field.dtype.kind == 'i':
                        field = field.astype("int")
                
                    if len(field) == shape[0]:
                        metadata[mkey] = field[i]
                    else:
                        metadata[mkey] = field
                else:
                    metadata[mkey] = field

            metadata['entry'] = entryname
            dataset = SansData(data=areaDetector[i].copy(), metadata=metadata)
            datasets.append(dataset)            

    return datasets




import copy

from .core import load_instrument
from .cache import get_cache
from . import fetch
from reductus.configurations import default

DEFAULT_CONFIG = copy.deepcopy(default.config)

def load_update(name="config_overrides"):
    """
    Load named configuration from reductus.configurations folder, and update
    a copy of default configuration with user settings
    """
    import importlib
    config_module = importlib.import_module("configurations.{name}".format(name=name))
    config_overrides = copy.deepcopy(config_module.config)
    config = copy.deepcopy(DEFAULT_CONFIG)
    config.update(config_overrides)
    return config

def load_config(name="config", fallback=True):
    """
    Look for configurations defined in the configurations directory
    if the name is not found, use "default_config" if fallback==True
    """
    import importlib

    try:
        config_module = importlib.import_module("configurations.{name}".format(name=name))
        return copy.deepcopy(config_module.config)
    except ImportError:
        if fallback:
            return DEFAULT_CONFIG
        else:
            raise

def apply_config(user_config=None, user_overrides=None):
    if user_config is not None:
        config = copy.deepcopy(user_config)
    else:
        config = copy.deepcopy(DEFAULT_CONFIG)

    if user_overrides is not None:
        config.update(user_overrides)

    fetch.DATA_SOURCES = config.get("data_sources", [])
    fetch.FILE_HELPERS = {
        source["name"]: source.get("file_helper_url", None)
        for source in fetch.DATA_SOURCES}

    cache_config = config.get('cache', False)
    if cache_config:
        cache_engine = cache_config.get("engine", None)
        cache_params = cache_config.get("params", {})
        cache_compression = cache_config.get("compression", False)
        cache_manager = get_cache()
        if cache_engine == "diskcache":
            cache_manager.use_diskcache(**cache_params)
        elif cache_engine == "redis":
            cache_manager.use_redis(**cache_params)
        else:
            cache_manager.use_memory()

        cache_manager._use_compression = cache_compression

    # Load refl instrument if nothing specified in config.
    # Note: instrument names do not match instrument ids.
    instruments = config.get('instruments', ['refl'])
    for name in instruments:
        load_instrument(name)

    # this is a strange thing to have here... but when connected
    # through a firewall that doesn't resolve IPV6 addresses correctly,
    # the application slows down tremendously.  There is no built-in
    # way to ask requests to use IPV4, so this is the recipe found:
    # https://stackoverflow.com/questions/33046733/force-requests-to-use-ipv4-ipv6/46972341#46972341
    if config.get('force_IPV4', False):
        import socket
        import requests.packages.urllib3.util.connection as urllib3_cn

        def allowed_gai_family():
            family = socket.AF_INET    # force IPv4
            return family

        urllib3_cn.allowed_gai_family = allowed_gai_family



# This program is in the public domain
"""
Join reflectivity datasets with matching intent/cross section.
"""
from __future__ import print_function

from copy import deepcopy

import numpy as np

from reductus.dataflow.lib import unit
from .refldata import Intent, ReflData, Environment
from .util import poisson_average, extend
from .resolution import divergence_simple, dTdL2dQ, TiTdL2Qxz

try:
    #from typing import List, Dict, Union, Sequence
    #Columns = Dict[str, List[np.ndarray]]
    #StackedColumns = Dict[str, np.ndarray]
    #IndexSet = List[int]
    pass
except ImportError:
    pass

def sort_files(datasets, key):
    # type: (List[ReflData], str) -> List[ReflData]
    """
    Order files by key.

    key can be one of: file, time, theta, or slit
    """
    if key == 'file':
        keyfn = lambda data: data.name
    elif key == 'time':
        import datetime
        keyfn = lambda data: (data.date + datetime.timedelta(seconds=data.monitor.start_time[0]))
    elif key == "theta":
        keyfn = lambda data: (data.sample.angle_x[0], data.detector.angle_x[0])
    elif key == "slit":
        keyfn = lambda data: (data.slit1.x, data.slit2.x)
    elif key == "none":
        return datasets
    else:
        raise ValueError("Unknown sort key %r: use file, time, theta or slit"
                         % key)
    datasets = datasets[:]
    datasets.sort(key=keyfn)
    return datasets


def join_datasets(group, Qtol, dQtol, by_Q=False):
    # type: (List[ReflData], float, float, bool) -> ReflData
    """
    Create a new dataset which joins the results of all datasets in the group.

    This is a multistep operation with the various parts broken into separate
    functions.
    """
    #print "joining files in",group[0].path,group[0].name,group[0].entry
    # Make sure all datasets are normalized by the same factor.
    normbase = group[0].normbase
    assert all(data.normbase == normbase for data in group), "can't mix time and monitor normalized data"

    # Gather the columns
    fields = get_fields(group)
    env = get_env(group)
    fields.update(env)
    columns = stack_columns(group, fields)
    columns = apply_mask(group, columns)
    columns = set_QdQ(columns)

    # Group points together, either by target angles, by actual angles or by Q
    # TODO: maybe include target Q
    # TODO: check background subtraction based on trajectoryData._q
    # ----- it can't be right since Qz_target is not properly propagated
    # ----- through the join...
    if Qtol == 0. and dQtol == 0.:
        targets = get_target_values(group)
        targets = stack_columns(group, targets)
        targets = apply_mask(group, targets)
        groups = group_by_target_angles(targets)
    elif by_Q:
        groups = group_by_Q(columns, Qtol=Qtol, dQtol=dQtol)
    else:
        groups = group_by_actual_angles(columns, Qtol=Qtol, dQtol=dQtol)

    # Join the data points in the individual columns
    columns = merge_points(groups, columns, normbase)

    # Sort so that points are in display order
    # Column keys are:
    #    Qx, Qz, dQ: Q and resolution
    #    Td: detector theta
    #    Ti: incident (sample) theta
    #    dT: angular divergence
    #    Li: monochromator wavelength
    #    Ld: detector wavelength
    #    dL: wavelength dispersion
    isslit = Intent.isslit(group[0].intent)
    isrock = Intent.isrock(group[0].intent)
    if isrock:
        # Sort detector rocking curves so that small deviations in sample
        # angle don't throw off the order in detector angle.
        keys = ('Qx', 'Qz', 'dQ')
        #keys = ('Td', 'Ti', 'Ld', 'dT', 'dL')
    elif isslit:
        keys = ('dT', 'Ld', 'dL')
    else:
        keys = ('Qz', 'dQ', 'Qx')
        #keys = ('Ti', 'Td', 'Ld', 'dT', 'dL')
    columns = sort_columns(columns, keys)

    data = build_dataset(group, columns, normbase)
    #print "joined",data.intent
    return data


def build_dataset(group, columns, norm):
    # type: (List[ReflData], StackedColumns, str) -> ReflData
    """
    Build a new dataset from a set of columns.

    Metadata is set from the first dataset in the group.

    If there are any sample environment columns they will be added to
    data.sample.environment.
    """
    head = group[0]

    # Copy details of first file as metadata for the returned dataset.
    # Note: using deepcopy since this is going to update subgroup
    # data such as data.slit1.x.
    data = deepcopy(group[0])
    ## Could instead do a semi-deep copy using info from the group:
    #data = copy(group[0])
    #for group_name, _ in data._groups:
    #    setattr(data, group_name, copy(getattr(data, group_name)))

    # Clear the fields that are no longer defined
    data.sample.angle_y = None
    data.sample.rotation = None
    data.detector.angle_y = None
    data.detector.rotation = None
    for k in range(1, 5):
        slit = getattr(data, 'slit%d'%k)
        slit.x = slit.y = slit.x_target = slit.y_target = None

    # summary data derived from group, or copied from head if commented
    #data.instrument
    #data.geometry
    #data.probe
    data.path = None  # no longer refers to head file
    data.uri = None  # TODO: does a reduction have a DOI?
    data.points = len(columns['v'])
    #data.channels  # unused
    #data.scale = 1.0  # unused
    #data.name
    #data.entry
    #data.description
    data.date = min(d.date for d in group)  # set date to earliest date
    data.duration = sum(d.duration for d in group)  # cumulative duration
    data.polarization = (head.polarization
                         if all(d.polarization == head.polarization for d in group)
                         else '')
    #data.normbase
    data.warnings = []  # initialize per-file history
    #data.vlabel
    #data.vunits
    #data.vscale
    #data.xlabel
    #data.xunits
    #data.xscale
    data.mask = None  # all points are active after join
    #data.angular_resolution # averaged
    data.Qz_basis = head.Qz_basis
    data.scan_value = []  # TODO: may want Td, Ti as alternate scan axes
    data.scan_label = []
    data.scan_units = []
    data.intent = head.intent  # preserve intent
    #data.x  # read-only
    #data.dx # read-only
    #data.v  # averaged
    #data.dv # averaged
    #data.Qz # read-only  # TODO: if join by Q then Qx,Qz,dQ need to be set
    #data.Qx # read-only
    #data.dQ # read-only

    # Fill in rates and monitors.
    # Note: we are not tracking monitor variance, so assume it is equal
    # to monitor counts (dm^2 = m).  This does not account for monitor
    # scaling due to dead time correction, etc.  In practice it doesn't
    # matter since we've already normalized the counts to a count rate
    # and we don't need detector counts or variance.
    v, dv, m, t = columns['v'], columns['dv'], columns['monitor'], columns['time']
    dmsq = m
    data.v = v
    data.dv = dv
    data.monitor.count_time = t
    data.monitor.counts = m
    data.monitor.counts_variance = dmsq
    data.monitor.roi_counts = columns['roi']
    data.monitor.roi_variance = columns['roi']
    data.monitor.source_power = columns['source_power']
    data.monitor.source_power_variance = columns['source_power_variance']


    # Assign a value to detector counts. We need this if we norm after join.
    if norm == "none":
        # v = counts, dv = dcounts
        data.detector.counts = v
        data.detector.counts_variance = dv**2
    elif norm == "time":
        # v = counts/time, dv = dcounts/time
        data.detector.counts = v * extend(t, v)
        data.detector.counts_variance = (dv * extend(t, dv))**2
    elif norm == "monitor":
        # v = counts/monitor, (dv/v)^2 = (dcounts/counts)^2+(dmonitor/monitor)^2
        # => dc^2 = (m dv)^2 - (v dm)^2
        data.detector.counts = v * extend(m, v)
        data.detector.counts_variance = (extend(m, dv)*dv)**2 - v**2*extend(dmsq,v)

    # Fill in the fields we have averaged
    data.sample.angle_x = columns['Ti']
    data.detector.angle_x = columns['Td']
    data.sample.angle_x_target = columns['Ti_target']
    data.detector.angle_x_target = columns['Td_target']
    data.slit1.x = columns['s1']
    data.slit2.x = columns['s2']
    data.slit3.x = columns['s3']
    data.slit4.x = columns['s4']
    # TODO: cleaner handling of candor data
    # Angular resolution may be stored separately from dT in the joined set
    # if it is multidimensional or dT is set to something else for grouping.
    res = 'angular_resolution' if 'angular_resolution' in columns else 'dT'
    data.angular_resolution = columns[res]
    # Some fields may not have been in the original data
    if data.Qz_target is not None:
        data.Qz_target = columns['Qz_target']
    if data.monochromator.wavelength is not None:
        data.monochromator.wavelength = columns['Li']
    if data.detector.wavelength is not None:
        # On candor data.detector.wavelength has shape [1, 2, 54] since it is
        # constant for all measurement points. Since Ld and dL need to be
        # scalars for grouping it is inconvenient to maintain the
        # full wavelength for each frame, so we instead assume that any time
        # the wavelength is multidimensional then it is constant. Further,
        # we assume that the constant is included in the group[0] metadata
        # we use as the basis of our return value.
        if getattr(data.detector.wavelength, 'ndim', 0) < 2:
            data.detector.wavelength = columns['Ld']
            data.detector.wavelength_resolution = columns['dL']

    # Add in any sample environment fields
    data.sample.environment = {}
    for k, v in head.sample.environment.items():
        if k in columns:
            env = Environment()
            env.units = v.units
            env.average = columns[k]
            data.sample.enviroment[k] = env
            # TODO: could maybe join the environment logs
            # ----- this would require setting them all to a common start time

    return data


def get_fields(group):
    # type: (List[ReflData]) -> Columns
    """
    Extract geometry and counts from all files in group into separate fields.

    Returns a map of columns: list of vectors, with one vector for each
    dataset in the group.
    """
    columns = dict(
        s1=[data.slit1.x for data in group],
        s2=[data.slit2.x for data in group],
        s3=[data.slit3.x for data in group],
        s4=[data.slit4.x for data in group],
        Ti=[data.sample.angle_x for data in group],
        Td=[data.detector.angle_x for data in group],
        dT=[data.angular_resolution for data in group],
        Li=[data.monochromator.wavelength for data in group],
        Ld=[data.detector.wavelength for data in group],
        dL=[data.detector.wavelength_resolution for data in group],
        monitor=[data.monitor.counts for data in group],
        roi=[data.monitor.roi_counts for data in group],
        source_power=[data.monitor.source_power for data in group],
        source_power_variance=[data.monitor.source_power_variance for data in group],
        time=[data.monitor.count_time for data in group],
        Ti_target=[data.sample.angle_x_target for data in group],
        Td_target=[data.detector.angle_x_target for data in group],
        Qz_target=[data.Qz_target for data in group],
        # using v,dv since poisson average wants rates
        v=[data.v for data in group],
        dv=[data.dv for data in group],
    )
    # TODO: cleaner way of handling multi-channel detectors
    if columns['v'][0].ndim > 1:
        # For candor analysis (and anything else with an nD detector), pick
        # a particular detector channel to use as the basis for merging
        # detector frames. This has only been tested on Candor so far, and
        # may need to change if we want to handle overlapping frames in a
        # traditional detector with a psd.

        # Save the angular divergence since it may be multidimensional, and
        # we will want to paste it in to the joined dataset. We will replace
        # dT with a scalar for the purposes of grouping frames.
        columns['angular_resolution'] = columns['dT']

        # Make dT a scalar. Use S1 rather than dT since candor does not yet
        # compute divergence properly.  Otherwise use the first channel of dT.
        columns['dT'] = columns['s1']
        #columns['dT'] = [v.flat[0] for v in columns['dT']]

        # Pick the first channel of L, dL.  Not picking a column since they are
        # constant across all frames, and since the column stacker automatically
        # extends a scalar to a vector with one entry per frame.
        columns['Ld'] = [v.flat[0] for v in columns['Ld']]
        columns['dL'] = [v.flat[0] for v in columns['dL']]

    return columns


def get_env(group):
    # type: (List[ReflData]) -> Columns
    """
    Extract sample environment from all fields in group into separate fields.
    """
    head = group[0]
    # Gather environment variables such as temperature and field.
    # Make sure they are all in the same units.
    columns = dict((e.name, []) for e in head.sample.environment)
    converter = dict((e.name, unit.Converter(e.units)) for e in head.sample.environment)
    for data in group:
        for env_name, env_list in columns.items():
            env = data.sample.environment.get(env_name, None)
            if env is not None:
                values = converter[env_name](env.average, units=env.units)
            else:
                values = None
            env_list.append(values)

    # Drop environment variables that are not defined in every file
    columns = dict((env_name, env_list)
                   for env_name, env_list in columns.items()
                   if not any(v is None for v in env_list))
    return columns


def stack_columns(group, columns):
    # type: (List[ReflData], Columns) -> StackedColumns
    """
    Join individual datasets into only long vector for each field in columns.
    """
    columns = dict((field, [_scalar_to_vector(part, data, field)
                            for part, data in zip(values, group)])
                   for field, values in columns.items())

    # Turn the data into arrays.
    #for k, v in columns.items(): print(f"{k}:", [vk.shape for vk in v])
    columns = dict((k, np.concatenate(v, axis=0)) for k, v in columns.items())
    return columns


def _scalar_to_vector(value, data, field):
    # type: (Union[np.ndarray, float], ReflData, str) -> np.ndarray
    """
    Make v a vector of length n if v is a scalar, or leave it alone.
    """
    n = len(data.v)
    if value is None:  # Convert missing data to NaN
        value = np.nan
    if np.isscalar(value) or len(value) == 1:
        return np.repeat(value, n, axis=0)
    elif len(value) == n:
        return value
    else:
        raise ValueError("%s length does not match data length in %s%s"
                         % (field, data.name, data.polarization))


def apply_mask(group, columns):
    # type: (List[ReflData], StackedColumns) -> StackedColumns
    """
    Mask out selected points from the joined dataset.

    Note: could instead return the non-masked points as the initial index set.
    """
    masks = [data.mask for data in group]
    if any(mask is not None for mask in masks):
        masks = [(data.mask&np.isfinite(data.v) if data.mask is not None else np.isfinite(data.v))
                 for data in group]
        idx = np.hstack(masks)
        columns = dict((k, v[idx]) for k, v in columns.items())
    return columns


def set_QdQ(columns):
    # type: (StackedColumns) -> StackedColumns
    """
    Generate Q and dQ fields from angles and geometry
    """
    Ti, Td, dT = columns['Ti'], columns['Td'], columns['dT']
    Ld, dL = columns['Ld'], columns['dL']
    #print("set_QdQ", [v.shape for v in (Ti, Td, dT, Ld, dL)])
    Qx, Qz = TiTdL2Qxz(Ti, Td, Ld)
    # TODO: is dQx == dQz ?
    dQ = dTdL2dQ(Td-Ti, dT, Ld, dL)
    columns['Qx'], columns['Qz'], columns['dQ'] = Qx, Qz, dQ
    return columns


def get_target_values(group):
    # type: (List[ReflData]) -> Columns
    """
    Get the target values for the instrument geometry.
    """
    columns = dict(
        Ti=[data.sample.angle_x_target for data in group],
        Td=[data.detector.angle_x_target for data in group],
        dT=[_target_dT(data) for data in group],
        Ld=[data.detector.wavelength for data in group],
        dL=[data.detector.wavelength_resolution for data in group],
    )
    return columns


def _target_dT(data):
    # type: (ReflData) -> np.ndarray
    """
    Idealized resolution based on the resolution of the target slits.
    """
    distance = abs(data.slit1.distance), abs(data.slit2.distance)
    slits = data.slit1.x_target, data.slit2.x_target
    return divergence_simple(slits=slits, distance=distance, use_sample=False)


def group_by_target_angles(columns):
    # type: (StackedColumns) -> List[IndexSet]
    """
    Given columns of target values, group together exactly matching points.
    """
    Ti, Td, dT = columns['Ti'], columns['Td'], columns['dT']
    Ld, dL = columns['Ld'], columns['dL']
    points = {}
    for index, point in enumerate(zip(Ti, Td, dT, Ld, dL)):
        points.setdefault(point, []).append(index)
    return list(points.values())


def group_by_actual_angles(columns, Qtol, dQtol):
    # type: (StackedColumns, float, float) -> List[IndexSet]
    """
    Given instrument geometry columns group points by angles and wavelength.
    """
    Ti, Td, dT = columns['Ti'], columns['Td'], columns['dT']
    Ld, dL = columns['Ld'], columns['dL']
    #print "joining", Qtol, dQtol, Ti, Td, dT
    groups = [list(range(len(Ti)))]
    groups = _group_by_dim(groups, Td, Qtol*dT)
    #print("Td groups", groups)
    groups = _group_by_dim(groups, Ti, Qtol*dT)
    #print("Ti groups", groups)
    groups = _group_by_dim(groups, Ld, Qtol*dL)
    #print("Ld groups", groups)
    groups = _group_by_dim(groups, dT, dQtol*dT)
    #print("dT groups", groups)
    groups = _group_by_dim(groups, dL, dQtol*dL)
    #print("dL groups", groups)
    return groups


def group_by_Q(columns, Qtol, dQtol):
    # type: (StackedColumns, float, float) -> List[IndexSet]
    """
    Given instrument geometry columns group points by Q and resolution.
    """
    Qx, Qz, dQ = columns['Qx'], columns['Qz'], columns['dQ']
    groups = [list(range(len(Qz)))]
    groups = _group_by_dim(groups, dQ, dQtol*dQ)
    groups = _group_by_dim(groups, Qz, Qtol*dQ)
    groups = _group_by_dim(groups, Qx, Qtol*dQ)
    return groups


def _group_by_dim(index_sets, data, width):
    # type: (List[IndexSet], np.ndarray, np.ndarray) -> List[IndexSet]
    """
    Given a list of index groups, split each subgroup according to the
    dimension given in data, making sure points in the group lie within
    width of each other.

    Note that the resolution dimensions must be split before the angle
    and wavelength dimensions, otherwise there can be weirdness. When points
    with widely different resolution are joined, there will be a new output
    group triggered every time a point with tight resolution is encountered,
    splitting up a series of points with loose resolution that would otherwise
    be joined.
    """
    refinement = []
    for subgroup in index_sets:
        refinement.extend(_split_subgroup(subgroup, data, width))
    return refinement


def _split_subgroup(indices, data, width):
    # type: (IndexSet, np.ndarray, np.ndarray) -> List[IndexSet]
    """
    Split an index group according to data, returning a list of subgroups.
    """

    # If there is only one point then there is nothing to split
    if len(indices) <= 1:
        return [indices]

    # Grab the data/width subset according to indices
    data = data[indices]
    width = width[indices]
    order = np.argsort(data)  # type: Sequence[int]

    # Initialize the returned group list and the current group; set start
    # and end points so that a new group is triggered on the first point
    groups = []  # type: List[IndexSet]
    current_group = []  # type: IndexSet
    start_point = end_point = -np.inf
    for k in order:
        # Check if next point falls out of bounds, either because it is
        # beyond the range of existing points (data[k] > end_point), or
        # because the first point is outside of the range the next
        # point (data[k]-width[k] > start_point).
        if data[k] > end_point or data[k] - width[k] > start_point:
            # next point is outside the range; close the current group and
            # start a new one.  The first time through the loop the current
            # group will be empty but still a new group will be triggered.
            if current_group:
                groups.append(current_group)
            current_group = [indices[k]]
            start_point = data[k]
            end_point = data[k] + width[k]
        else:
            # next point is in the range; add it to the current group and
            # limit the end point of the group to its range
            current_group.append(indices[k])
            end_point = min(end_point, data[k]+width[k])
    groups.append(current_group)
    return groups


def merge_points(index_sets, columns, normbase):
    # type: (List[IndexSet], StackedColumns, str) -> StackedColumns
    """
    Join points together according to groups.

    Points are weighted according to normbase, which could be 'monitor'
    or 'time'.

    Note: we do not yet increase divergence when points with slightly
    different incident angles are mixed.
    """
    # Weight each point by monitor/time/counts
    if normbase == "none":
        # if weighting by counts then use the counts across the entire
        # detector as the weight.  This is a just a proxy for time/monitor
        # weighting but using the measured data, assuming the same conditions
        # give the same count rate.
        counts = columns['v']
        ndim = counts[0].ndim
        if ndim == 1:
            weight = counts
        else:
            axis = tuple(range(1, ndim))
            weight = [np.sum(v, axis=axis) for v in counts]
    else:
        weight = columns[normbase]

    # build a structure to hold the results
    results = dict((k, []) for k in columns.keys())

    #for k,v in columns.items(): print k, len(v), v
    for group in index_sets:
        if len(group) == 1:
            index = group[0]
            for key, value in columns.items():
                results[key].append(value[index])
        else:
            v, dv = poisson_average(
                columns['v'][group], columns['dv'][group], norm=normbase)
            results['v'].append(v)
            results['dv'].append(dv)
            results['time'].append(np.sum(columns['time'][group]))
            results['monitor'].append(np.sum(columns['monitor'][group]))
            # TODO: dQ should increase when points are mixed (see MERGE below)
            w = weight[group]
            for key, value in columns.items():
                if key not in ['v', 'dv', 'time', 'monitor']:
                    results[key].append(np.average(value[group], weights=w, axis=0))

    # Turn lists into arrays
    results = dict((k, np.array(v)) for k, v in results.items())
    return results

# MERGE variance
#
# There is a simple expression for the moments of a mixture of distributions:
#     T = (sum wk . T) / (sum wk)
#     dT^2 = (sum wk . (dTk^2 + Tk^2)) / (sum wk) - T^2
# See: https://en.wikipedia.org/wiki/Mixture_density#Moments
#
# This formula should be applied to angular distribution since that is
# the quantity being mixed when combining measurements at slightly
# different angles.

def sort_columns(columns, keys):
    # type: (StackedColumns, Sequence[str]) -> StackedColumns
    """
    Returns the set of columns by a ordered by a list of keys.

    *columns* is a dictionary of vectors of the same length.

    *keys* is the list of columns in the order they should be sorted.
    """
    A = [columns[name] for name in reversed(keys)]
    A = np.array(A)
    index = np.lexsort(A)

    return dict((k, v[index]) for k, v in columns.items())


def demo():
    import sys
    import matplotlib.pyplot as plt
    from . import steps
    from .load import setup_fetch, fetch_uri
    from .util import group_by_key
    if len(sys.argv) == 1:
        print("usage: python -m reflred.joindata file...")
        sys.exit(1)
    np.set_printoptions(linewidth=10000)
    setup_fetch()
    data = []
    for uri in sys.argv[1:]:
        try:
            entries = fetch_uri(uri)
        except Exception as exc:
            print(str(exc) + " while loading " + uri)
            continue
        for entry in entries:
            entry = steps.mark_intent(steps.normalize(steps.divergence(entry)))
            if not Intent.isspec(entry.intent):
                continue
            #entry.plot()
            data.append(entry)
    groups = group_by_key('polarization', data).values()
    output = []
    for group in groups:
        group = sort_files(group, 'file')
        result = join_datasets(group, Qtol=0.5, dQtol=0.002)
        result.plot()
        output.append(result)
    plt.legend()
    plt.show()

if __name__ == "__main__":
    demo()



from __future__ import print_function

from os import listdir
from os.path import isfile, join as joinpath, dirname, abspath, exists

import regression

try:
    import pytest
    parametrize = pytest.mark.parametrize
except ImportError:
    # If pytest not installed then do simplet test suite which stops at the
    # first regression failure
    def parametrize(arg_names, values):
        def wrapper(function):
            def combined_tests():
                for test in values:
                    function(*test)
            return combined_tests
        return wrapper

# Prefer not to scan regression directory on module load, but pytest doesn't
# seem to have an interface for building tests on demand.
def get_regression_files():
    REGRESSION_PATH = abspath(joinpath(dirname(__file__), 'regression_files'))
    if exists(REGRESSION_PATH):
        data_files = (
            path
            for f in listdir(REGRESSION_PATH)
            # CRUFT: assignment expressions require python 3.8
            #if isfile(path := joinpath(REGRESSION_PATH, f))
            for path in [joinpath(REGRESSION_PATH, f)]
            if isfile(path))
    else:
        data_files = ()
    return data_files

@parametrize("path", get_regression_files())
def test_regression(path):
    #description = "Regression test on %s"%path
    regression.replay_file(path)



"""
Context manager for numpy random number generator (pre-1.17 interface).
"""

import numpy as np

# TODO: consider intercepting default_rng() and/or SeedSequence in np.random
#
# For code which goes to the trouble of maintaining its own sequences with
# the post-1.17 np.random interface, the current context manager will not work.
# We can get a little further by monkey-patching np.random with our own
# default_rng/SeedSequence in which we control the entropy, but this won't
# help with modules that use "from numpy.random import default_rng".  Maybe
# with some really deep python trickery (replacing the content of one
# PyObject with another in memory would be possible in cpython with some
# very unsafe code) ...

class push_seed(object):
    """
    Set the seed value for the random number generator.

    When used in a with statement, the random number generator state is
    restored after the with statement is complete.

    Note: the documentation for numpy.random has declared the pre-1.17
    interface as a legacy api, and is encouraging new code to create
    and manage their own random number streams. These streams will
    not be affected by the push_seed interface, so you will need to do
    something else to guarantee reproducible simulations with such code.

    :Parameters:

    *seed* : int or array_like, optional
        Seed for RandomState

    :Example:

    Seed can be used directly to set the seed::

        >>> from numpy.random import randint
        >>> push_seed(24)
        <...push_seed object at...>
        >>> print(randint(0,1000000,3))
        [242082    899 211136]

    Seed can also be used in a with statement, which sets the random
    number generator state for the enclosed computations and restores
    it to the previous state on completion::

        >>> with push_seed(24):
        ...    print(randint(0,1000000,3))
        [242082    899 211136]

    Using nested contexts, we can demonstrate that state is indeed
    restored after the block completes::

        >>> with push_seed(24):
        ...    print(randint(0,1000000))
        ...    with push_seed(24):
        ...        print(randint(0,1000000,3))
        ...    print(randint(0,1000000))
        242082
        [242082    899 211136]
        899

    The restore step is protected against exceptions in the block::

        >>> with push_seed(24):
        ...    print(randint(0,1000000))
        ...    try:
        ...        with push_seed(24):
        ...            print(randint(0,1000000,3))
        ...            raise Exception()
        ...    except Exception:
        ...        print("Exception raised")
        ...    print(randint(0,1000000))
        242082
        [242082    899 211136]
        Exception raised
        899
    """
    def __init__(self, seed=None):
        self._state = np.random.get_state()
        np.random.seed(seed)

    def __enter__(self):
        return None

    def __exit__(self, *args):
        np.random.set_state(self._state)
